"""Module containing the `definitions`.

The `definitions` module provides a set of predefined types used in the package.

Classes:
    Entity: The class representing the anonipy entity object.
    Replacement: The class representing the anonipy replacement object.

"""

import re
from typing import Union, TypedDict
from typing_extensions import NotRequired
from dataclasses import dataclass

from .utils.regex import regex_mapping
from .constants import ENTITY_TYPES

# ================================================
# Entity Definitions
# ================================================


@dataclass
class Entity:
    """The class representing the anonipy Entity object.

    Attributes:
        text (str): The text of the entity.
        label (str): The label of the entity.
        start_index (int): The start index of the entity in the text.
        end_index (int): The end index of the entity in the text.
        score (float): The prediction score of the entity. The score is returned by the extractor models.
        type (ENTITY_TYPES): The type of the entity.
        regex (Union[str, re.Pattern]): The regular expression the entity must match.

    """

    text: str
    label: str
    start_index: int
    end_index: int
    score: float = 1.0
    type: ENTITY_TYPES = None
    regex: Union[str, re.Pattern] = None

    def __post_init__(self):
        if self.regex is None:
            if self.type == "custom":
                raise ValueError("Custom entities require a regex.")
            self.regex = regex_mapping[self.type]

    def get_regex_group(self) -> Union[str, None]:
        """Get the regex group.

        Returns:
            The regex group.

        """

        p_match = re.match(r"^.*?\((.*)\).*$", self.regex)
        return p_match.group(1) if p_match else self.regex

    def __str__(self) -> str:
        """String representation of the entity.

        Returns:
            The string representation of the entity.

        """
        return f"Entity(text='{self.text}', label='{self.label}', start_index={self.start_index}, end_index={self.end_index}, type='{self.type}')"


class Replacement(TypedDict):
    """The class representing the anonipy Replacement object.

    Attributes:
        original_text (str): The original text of the entity.
        label (str): The label of the entity.
        start_index (int): The start index of the entity in the text.
        end_index (int): The end index of the entity in the text.
        anonymized_text (str): The anonymized text replacing the original.

    """

    original_text: NotRequired[str]
    label: NotRequired[str]
    start_index: int
    end_index: int
    anonymized_text: str



import re
import warnings
import importlib
from typing import List, Tuple

import torch
from spacy import displacy
from spacy.tokens import Doc, Span
from spacy.language import Language

from ..helpers import (
    convert_spacy_to_entity,
    detect_repeated_entities,
    get_doc_entity_spans,
    create_spacy_entities,
)
from ...utils.regex import regex_mapping
from ...constants import LANGUAGES
from ...definitions import Entity
from ...utils.colors import get_label_color

from .interface import ExtractorInterface


# ===============================================
# Extractor class
# ===============================================


class NERExtractor(ExtractorInterface):
    """The class representing the named entity recognition (NER) extractor.

    Examples:
        >>> from anonipy.constants import LANGUAGES
        >>> from anonipy.anonymize.extractors import NERExtractor
        >>> labels = [{"label": "PERSON", "type": "string"}]
        >>> extractor = NERExtractor(labels, lang=LANGUAGES.ENGLISH)
        >>> extractor("John Doe is a 19 year old software engineer.", detect_repeats=False)
        Doc, [Entity]

    Attributes:
        labels (List[dict]): The list of labels to extract.
        lang (str): The language of the text to extract.
        score_th (float): The score threshold.
        use_gpu (bool): Whether to use GPU.
        gliner_model (str): The gliner model to use.
        pipeline (Language): The spacy pipeline for extracting entities.
        spacy_style (str): The style the entities should be stored in the spacy doc.

    Methods:
        __call__(self, text):
            Extract the entities from the text.
        display(self, doc):
            Display the entities in the text.

    """

    def __init__(
        self,
        labels: List[dict],
        *args,
        lang: LANGUAGES = LANGUAGES.ENGLISH,
        score_th: float = 0.5,
        use_gpu: bool = False,
        gliner_model: str = "E3-JSI/gliner-multi-pii-domains-v1",
        spacy_style: str = "ent",
        **kwargs,
    ):
        """Initialize the named entity recognition (NER) extractor.

        Examples:
            >>> from anonipy.constants import LANGUAGES
            >>> from anonipy.anonymize.extractors import NERExtractor
            >>> labels = [{"label": "PERSON", "type": "string"}]
            >>> extractor = NERExtractor(labels, lang=LANGUAGES.ENGLISH)
            NERExtractor()

        Args:
            labels: The list of labels to extract.
            lang: The language of the text to extract.
            score_th: The score threshold. Entities with a score below this threshold will be ignored.
            use_gpu: Whether to use GPU.
            gliner_model: The gliner model to use to identify the entities.
            spacy_style: The style the entities should be stored in the spacy doc. Options: `ent` or `span`.

        """

        super().__init__(labels, *args, **kwargs)
        self.lang = lang
        self.score_th = score_th
        self.use_gpu = use_gpu
        self.gliner_model = gliner_model
        self.spacy_style = spacy_style
        self.labels = self._prepare_labels(labels)

        with warnings.catch_warnings():
            # TODO: remove once the GLiNER package includes the fix (inproper file closing)
            warnings.filterwarnings("ignore", category=ResourceWarning)
            self.pipeline = self._prepare_pipeline()

    def __call__(
        self, text: str, detect_repeats: bool = False, *args, **kwargs
    ) -> Tuple[Doc, List[Entity]]:
        """Extract the entities from the text.

        Examples:
            >>> extractor("John Doe is a 19 year old software engineer.", detect_repeats=False)
            Doc, [Entity]

        Args:
            text: The text to extract entities from.
            detect_repeats: Whether to check text again for repeated entities.

        Returns:
            The spacy document.
            The list of extracted entities.

        """

        doc = self.pipeline(text)
        anoni_entities, spacy_entities = self._prepare_entities(doc)

        if detect_repeats:
            anoni_entities = detect_repeated_entities(
                doc, anoni_entities, self.spacy_style
            )

        create_spacy_entities(doc, anoni_entities, self.spacy_style)

        return doc, anoni_entities

    def display(self, doc: Doc, page: bool = False, jupyter: bool = None) -> str:
        """Display the entities in the text.

        Examples:
            >>> doc, entities = extractor("John Doe is a 19 year old software engineer.")
            >>> extractor.display(doc)
            HTML

        Args:
            doc: The spacy doc to display.
            page: Whether to display the doc in a web browser.
            jupyter: Whether to display the doc in a jupyter notebook.

        Returns:
            The HTML representation of the document and the extracted entities.

        """

        options = {
            "colors": {l["label"]: get_label_color(l["label"]) for l in self.labels}
        }
        return displacy.render(
            doc, style=self.spacy_style, options=options, page=page, jupyter=jupyter
        )

    # ===========================================
    # Private methods
    # ===========================================

    def _prepare_labels(self, labels: List[dict]) -> List[dict]:
        """Prepare the labels for the extractor.

        The provided labels are enriched with the corresponding regex
        definitions, if the `regex` key was not provided.

        Args:
            labels: The list of labels to prepare.

        Returns:
            The enriched labels.

        """
        for l in labels:
            if "regex" in l:
                continue
            regex = regex_mapping[l["type"]]
            if regex is not None:
                l["regex"] = regex
        return labels

    def _create_gliner_config(self) -> dict:
        """Create the config for the GLINER model.

        Returns:
            The configuration dictionary for the GLINER model.

        """

        map_location = "cpu"
        if self.use_gpu and not torch.cuda.is_available():
            return warnings.warn(
                "The user requested GPU use, but not available GPU was found. Reverting back to CPU use."
            )
        if self.use_gpu and torch.cuda.is_available():
            map_location = "cuda"

        return {
            # the model is specialized for extracting PII data
            "gliner_model": self.gliner_model,
            "labels": [l["label"] for l in self.labels],
            "threshold": self.score_th,
            "chunk_size": 384,
            "style": self.spacy_style,
            "map_location": map_location,
        }

    def _prepare_pipeline(self) -> Language:
        """Prepare the spacy pipeline.

        Prepares the pipeline for processing the text in the corresponding
        provided language.

        Returns:
            The spacy text processing and extraction pipeline.

        """

        # load the appropriate parser for the language
        module_lang, class_lang = self.lang[0].lower(), self.lang[1].lower().title()
        language_module = importlib.import_module(f"spacy.lang.{module_lang}")
        language_class = getattr(language_module, class_lang)
        # initialize the language parser
        nlp = language_class()
        nlp.add_pipe("sentencizer")
        gliner_config = self._create_gliner_config()
        nlp.add_pipe("gliner_spacy", config=gliner_config)
        return nlp

    def _prepare_entities(self, doc: Doc) -> Tuple[List[Entity], List[Span]]:
        """Prepares the anonipy and spacy entities.

        Args:
            doc: The spacy doc to prepare.

        Returns:
            The list of anonipy entities.
            The list of spacy entities.

        """

        # TODO: make this part more generic
        anoni_entities = []
        spacy_entities = []
        for s in get_doc_entity_spans(doc, self.spacy_style):
            label = list(filter(lambda x: x["label"] == s.label_, self.labels))[0]
            if re.match(label["regex"], s.text):
                anoni_entities.append(convert_spacy_to_entity(s, **label))
                spacy_entities.append(s)
        return anoni_entities, spacy_entities



"""Module containing the `generators`.

The `generators` module provides a set of generators used to generate data
substitutes.

Classes:
    LLMLabelGenerator: The class representing the label generator utilizing LLMs.
    MaskLabelGenerator: The class representing the label generator utilizing token masking.
    NumberGenerator: The class representing the number generator.
    DateGenerator: The class representing the date generator.

"""

from .interface import GeneratorInterface
from .llm_label_generator import LLMLabelGenerator
from .mask_label_generator import MaskLabelGenerator
from .number_generator import NumberGenerator
from .date_generator import DateGenerator

__all__ = [
    "LLMLabelGenerator",
    "MaskLabelGenerator",
    "NumberGenerator",
    "DateGenerator",
    "GeneratorInterface",
]



from typing import List, Tuple

from spacy.tokens import Doc
from ...definitions import Entity


class ExtractorInterface:
    """The class representing the extractor interface.

    All extractors should inherit from this class.

    Methods:
        __call__(text):
            Extract entities from the text.

    """

    def __init__(self, labels: List[dict], *args, **kwargs):
        pass

    def __call__(self, text: str, *args, **kwargs) -> Tuple[Doc, List[Entity]]:
        pass



"""Module containing the `extractors`.

The `extractors` module provides a set of extractors used to identify relevant
information within a document.

Classes:
    NERExtractor: The class representing the named entity recognition (NER) extractor.
    PatternExtractor: The class representing the pattern extractor.
    MultiExtractor: The class representing the multi extractor.

"""

from .interface import ExtractorInterface
from .multi_extractor import MultiExtractor
from .ner_extractor import NERExtractor
from .pattern_extractor import PatternExtractor


__all__ = ["ExtractorInterface", "MultiExtractor", "NERExtractor", "PatternExtractor"]



from typing import List, Tuple, Callable

from .interface import StrategyInterface
from ...definitions import Entity, Replacement
from ..helpers import anonymize

# =====================================
# Main class
# =====================================


class PseudonymizationStrategy(StrategyInterface):
    """The class representing the pseudonymization strategy.

    Examples:
        >>> from anonipy.anonymize.strategies import PseudonymizationStrategy
        >>> strategy = PseudonymizationStrategy(mapping)
        >>> strategy.anonymize(text, entities)

    Attributes:
        mapping: The mapping of entities to pseudonyms.

    Methods:
        anonymize(text, entities):
            Anonymize the text based on the entities.

    """

    def __init__(self, mapping: Callable, *args, **kwargs):
        """Initializes the pseudonymization strategy.

        Examples:
            >>> from anonipy.anonymize.strategies import PseudonymizationStrategy
            >>> strategy = PseudonymizationStrategy(mapping)

        Args:
            mapping: The mapping function on how to handle each entity type.

        """

        super().__init__(*args, **kwargs)
        self.mapping = mapping

    def anonymize(
        self, text: str, entities: List[Entity], *args, **kwargs
    ) -> Tuple[str, List[Replacement]]:
        """Anonymize the text using the pseudonymization strategy.

        Examples:
            >>> from anonipy.anonymize.strategies import PseudonymizationStrategy
            >>> strategy = PseudonymizationStrategy(mapping)
            >>> strategy.anonymize(text, entities)

        Args:
            text: The text to anonymize.
            entities: The list of entities to anonymize.

        Returns:
            The anonymized text.
            The list of applied replacements.

        """

        replacements = []
        for ent in entities:
            replacement = self._create_replacement(ent, text, replacements)
            replacements.append(replacement)
        anonymized_text, replacements = anonymize(text, replacements)
        return anonymized_text, replacements

    # ===========================================
    # Private methods
    # ===========================================

    def _create_replacement(
        self, entity: Entity, text: str, replacements: List[dict]
    ) -> Replacement:
        """Creates a replacement for the entity.

        Args:
            entity: The entity to create the replacement for.
            text: The text to anonymize.
            replacements: The list of existing replacements.

        Returns:
            The created replacement.

        """

        # check if the replacement already exists
        anonymized_text = self._check_replacement(entity, replacements)
        # create a new replacement if it doesn't exist
        anonymized_text = (
            self.mapping(text, entity) if not anonymized_text else anonymized_text
        )
        return {
            "original_text": entity.text,
            "label": entity.label,
            "start_index": entity.start_index,
            "end_index": entity.end_index,
            "anonymized_text": anonymized_text,
        }

    def _check_replacement(
        self, entity: Entity, replacements: List[Replacement]
    ) -> str:
        """Checks if a suitable replacement already exists.

        Args:
            entity: The entity to check.
            replacements: The list of replacements.

        Returns:
            The anonymized text if the replacement already exists, None otherwise.

        """
        existing_replacement = list(
            filter(lambda x: x["original_text"] == entity.text, replacements)
        )
        return (
            existing_replacement[0]["anonymized_text"]
            if len(existing_replacement) > 0
            else None
        )



from anonipy.anonymize import anonymize

# =====================================
# Helper functions
# =====================================

test_text = "Test this string, and this test too!"
test_text_anonymized = "********* this TYPE, and this NEWTEST too!"
test_replacements = [
    {
        "original_text": "Test",
        "label": "test",
        "start_index": 0,
        "end_index": 4,
        "anonymized_text": "*********",
    },
    {
        "start_index": 10,
        "end_index": 16,
        "anonymized_text": "TYPE",
    },
    {
        "original_text": "test",
        "label": "test",
        "start_index": 27,
        "end_index": 31,
        "anonymized_text": "NEWTEST",
    },
]

# =====================================
# Test Anonymize
# =====================================


def test_anonymize():
    anonymized_text, replacements = anonymize(test_text, test_replacements)
    assert anonymized_text == test_text_anonymized
    assert replacements == test_replacements



import pytest

from anonipy.utils.file_system import open_file
from test.resources.example_outputs import WORD_TEXT, PDF_TEXT, TXT_TEXT

RESOURCES = {
    "word": "./test/resources/example.docx",
    "pdf": "./test/resources/example.pdf",
    "txt": "./test/resources/example.txt",
}


@pytest.mark.parametrize(
    "file_type, expected_output",
    [
        ("word", WORD_TEXT),
        ("pdf", PDF_TEXT),
        ("txt", TXT_TEXT),
    ],
)
def test_open_file(file_type, expected_output):
    assert open_file(RESOURCES[file_type]) == expected_output



import pytest

from anonipy.utils.regex import (
    regex_mapping,
    RegexMapping,
    REGEX_STRING,
    REGEX_INTEGER,
    REGEX_FLOAT,
    REGEX_DATE,
    REGEX_EMAIL_ADDRESS,
    REGEX_PHONE_NUMBER,
    REGEX_WEBSITE_URL,
)
from anonipy.constants import ENTITY_TYPES

# =====================================
# Test Cases
# =====================================

TEST_CASES = [
    {"value": "string", "entity": ENTITY_TYPES.STRING, "regex": REGEX_STRING},
    {
        "value": "integer",
        "entity": ENTITY_TYPES.INTEGER,
        "regex": REGEX_INTEGER,
    },
    {"value": "float", "entity": ENTITY_TYPES.FLOAT, "regex": REGEX_FLOAT},
    {"value": "date", "entity": ENTITY_TYPES.DATE, "regex": REGEX_DATE},
    {
        "value": "email",
        "entity": ENTITY_TYPES.EMAIL,
        "regex": REGEX_EMAIL_ADDRESS,
    },
    {
        "value": "phone_number",
        "entity": ENTITY_TYPES.PHONE_NUMBER,
        "regex": REGEX_PHONE_NUMBER,
    },
    {
        "value": "website_url",
        "entity": ENTITY_TYPES.WEBSITE_URL,
        "regex": REGEX_WEBSITE_URL,
    },
    {"value": "custom", "entity": ENTITY_TYPES.CUSTOM, "regex": ".*"},
    {"value": "test", "entity": "test", "regex": ".*"},
]

# =====================================
# Test Entity
# =====================================


def test_init():
    assert isinstance(regex_mapping, RegexMapping)
    assert hasattr(regex_mapping, "regex_mapping")


@pytest.mark.parametrize("test_case", TEST_CASES)
def test_regex_mapping(test_case):
    assert regex_mapping[test_case["entity"]] == test_case["regex"]
    assert regex_mapping[test_case["value"]] == regex_mapping[test_case["entity"]]



import os
import shutil
import warnings

import pytest
from transformers import logging

from anonipy.anonymize.pipeline import Pipeline
from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
from anonipy.anonymize.strategies import RedactionStrategy
from anonipy.constants import LANGUAGES

# disable transformers logging
logging.set_verbosity_error()

# =====================================
# Test Pipeline
# =====================================


@pytest.fixture(autouse=True)
def suppress_warnings():
    warnings.filterwarnings("ignore", category=UserWarning)


@pytest.fixture(scope="module")
def setup():
    ner_labels = [{"label": "PERSON", "type": "string"}]
    pattern_labels = [{"label": "DATE", "type": "regex", "regex": r"\d{4}-\d{2}-\d{2}"}]
    extractors = [
        NERExtractor(ner_labels, lang=LANGUAGES.ENGLISH),
        PatternExtractor(pattern_labels, lang=LANGUAGES.ENGLISH),
    ]
    multi_extractor = MultiExtractor(extractors)
    strategy = RedactionStrategy()
    input_dir = "test/resources"
    output_dir = "test/output"
    yield extractors, multi_extractor, strategy, input_dir, output_dir
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)


def test_init():
    with pytest.raises(TypeError):
        Pipeline()


def test_init_extractor_single(setup):
    extractors, _, strategy, _, _ = setup
    pipeline = Pipeline(extractors[0], strategy)
    assert isinstance(pipeline, Pipeline)


def test_init_extractor_list(setup):
    extractors, _, strategy, _, _ = setup
    pipeline = Pipeline(extractors, strategy)
    assert isinstance(pipeline, Pipeline)


def test_init_extractor_multi(setup):
    _, multi_extractor, strategy, _, _ = setup
    pipeline = Pipeline(multi_extractor, strategy)
    assert isinstance(pipeline, Pipeline)


def test_methods(setup):
    _, multi_extractor, strategy, _, _ = setup
    pipeline = Pipeline(multi_extractor, strategy)
    assert hasattr(pipeline, "anonymize")


def test_anonymize(setup):
    _, multi_extractor, strategy, input_dir, output_dir = setup
    pipeline = Pipeline(multi_extractor, strategy)
    pipeline.anonymize(input_dir, output_dir)

    assert os.path.exists(output_dir)
    for root, _, files in os.walk(output_dir):
        for file in files:
            with open(os.path.join(root, file), "r") as f:
                assert f.read()


def test_anonymize_flatten(setup):
    _, multi_extractor, strategy, input_dir, output_dir = setup
    pipeline = Pipeline(multi_extractor, strategy)
    pipeline.anonymize(input_dir, output_dir, flatten=True)

    assert os.path.exists(output_dir)
    for root, _, files in os.walk(output_dir):
        for file in files:
            with open(os.path.join(root, file), "r") as f:
                assert f.read()


def test_anonymize_invalid_input_dir(setup):
    _, multi_extractor, strategy, _, output_dir = setup
    pipeline = Pipeline(multi_extractor, strategy)
    with pytest.raises(ValueError):
        pipeline.anonymize("invalid", output_dir)


def test_anonymize_invalid_output_dir(setup):
    _, multi_extractor, strategy, input_dir, _ = setup
    pipeline = Pipeline(multi_extractor, strategy)
    with pytest.raises(ValueError):
        pipeline.anonymize(input_dir, input_dir)



import re
import datetime
import warnings
from typing import Tuple

import dateparser
from babel.dates import format_datetime, format_datetime

# =====================================
# Constants
# =====================================


POSSIBLE_FORMATS = [
    "yyyy-MM-dd HH:mm:ss",
    "dd-MM-yyyy HH:mm:ss",
    "MM-dd-yyyy HH:mm:ss",
    "yyyy/MM/dd HH:mm:ss",
    "dd/MM/yyyy HH:mm:ss",
    "MM/dd/yyyy HH:mm:ss",
    "yyyy.MM.dd HH:mm:ss",
    "dd.MM.yyyy HH:mm:ss",
    "d.M.yyyy HH:mm:ss",
    "MM.dd.yyyy HH:mm:ss",
    "yyyy-MM-dd HH:mm",
    "dd-MM-yyyy HH:mm",
    "MM-dd-yyyy HH:mm",
    "yyyy/MM/dd HH:mm",
    "dd/MM/yyyy HH:mm",
    "MM/dd/yyyy HH:mm",
    "yyyy.MM.dd HH:mm",
    "dd.MM.yyyy HH:mm",
    "d.M.yyyy HH:mm",
    "MM.dd.yyyy HH:mm",
    "yyyy-MM-dd",
    "dd-MM-yyyy",
    "MM-dd-yyyy",
    "yyyy/MM/dd",
    "dd/MM/yyyy",
    "MM/dd/yyyy",
    "yyyy.MM.dd",
    "dd.MM.yyyy",
    "d.M.yyyy",
    "MM.dd.yyyy",
    "d MMMM yyyy HH:mm:ss",
    "d MMMM yyyy HH:mm",
    "d MMMM yyyy",
    "d. MMMM yyyy HH:mm:ss",
    "d. MMMM yyyy HH:mm",
    "d. MMMM yyyy",
    "d MMM yyyy HH:mm:ss",
    "d MMM yyyy HH:mm",
    "d MMM yyyy",
    "d. MMM yyyy HH:mm:ss",
    "d. MMM yyyy HH:mm",
    "d. MMM yyyy",
    "EEE, d MMMM yyyy HH:mm:ss",
    "EEE, d MMMM yyyy HH:mm",
    "EEE, d MMMM yyyy",
    "EEE, d MMMM, yyyy HH:mm:ss",
    "EEE, d MMMM, yyyy HH:mm",
    "EEE, d MMMM, yyyy",
    "EEE, d. MMMM yyyy HH:mm:ss",
    "EEE, d. MMMM yyyy HH:mm",
    "EEE, d. MMMM yyyy",
    "EEE, d. MMMM, yyyy HH:mm:ss",
    "EEE, d. MMMM, yyyy HH:mm",
    "EEE, d. MMMM, yyyy",
    "EEE, MMMM d, yyyy HH:mm:ss",
    "EEE, MMMM d, yyyy HH:mm",
    "EEE, MMMM d, yyyy",
    "EEE, MMMM d yyyy HH:mm:ss",
    "EEE, MMMM d yyyy HH:mm",
    "EEE, MMMM d yyyy",
    "EEEE, d MMMM yyyy HH:mm:ss",
    "EEEE, d MMMM yyyy HH:mm",
    "EEEE, d MMMM yyyy",
    "EEEE, d MMMM, yyyy HH:mm:ss",
    "EEEE, d MMMM, yyyy HH:mm",
    "EEEE, d MMMM, yyyy",
    "EEEE, d. MMMM yyyy HH:mm:ss",
    "EEEE, d. MMMM yyyy HH:mm",
    "EEEE, d. MMMM yyyy",
    "EEEE, d. MMMM, yyyy HH:mm:ss",
    "EEEE, d. MMMM, yyyy HH:mm",
    "EEEE, d. MMMM, yyyy",
    "EEEE, MMMM d, yyyy HH:mm:ss",
    "EEEE, MMMM d, yyyy HH:mm",
    "EEEE, MMMM d, yyyy",
    "EEEE, MMMM d yyyy HH:mm:ss",
    "EEEE, MMMM d yyyy HH:mm",
    "EEEE, MMMM d yyyy",
    "d 'de' MMMM 'de' yyyy HH:mm:ss",
    "d 'de' MMMM 'de' yyyy HH:mm",
    "d 'de' MMMM 'de' yyyy",
    "EEEE, d 'de' MMMM 'de' yyyy HH:mm:ss",
    "EEEE, d 'de' MMMM 'de' yyyy HH:mm",
    "EEEE, d 'de' MMMM 'de' yyyy",
    "EEEE, MMMM d 'de' yyyy HH:mm:ss",
    "EEEE, MMMM d 'de' yyyy HH:mm",
    "EEEE, MMMM d 'de' yyyy",
    "MMM d, yyyy HH:mm:ss",
    "MMM d, yyyy HH:mm",
    "MMM d, yyyy",
    "d MMM yyyy HH:mm:ss",
    "d MMM yyyy HH:mm",
    "d MMM yyyy",
    "EEE, MMM d, yyyy HH:mm:ss",
    "EEE, MMM d, yyyy HH:mm",
    "EEE, MMM d, yyyy",
]


# =====================================
# Auto date format detector
# =====================================


def detect_datetime_format(datetime: str, lang: str) -> Tuple[datetime.datetime, str]:
    """Detects the datetime format.

    Args:
        datetime: The datetime string to detect the format.
        lang: The language of the datetime string.

    Returns:
        The detected datetime and it's format.

    """

    fdatetime = _prepare_datetime(datetime, lang)

    try:
        # Suppress all DeprecationWarnings from dateparser's internal strptime usage
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)
            parsed_datetime = dateparser.parse(fdatetime, languages=[lang])

        for FMT in POSSIBLE_FORMATS:
            try:
                formatted_date = format_datetime(
                    parsed_datetime, format=FMT, locale=lang
                )
                if formatted_date == fdatetime:
                    return parsed_datetime, FMT
            except ValueError:
                continue

        warnings.warn(
            f"Could not detect the datetime format for `{datetime}`. Defaulting to `yyyy-MM-dd`..."
        )
        return parsed_datetime, "yyyy-MM-dd"

    except dateparser.ParserError:
        return None, None


def _prepare_datetime(datetime: str, lang: str) -> str:
    """Preares the datetime string for formatting.

    Args:
        datetime: The datetime string to format.
        lang: The language of the datetime string.

    Returns:
        The formatted datetime string.

    """

    if lang not in ["en", "de", "el"]:
        datetime = datetime.lower()

    # Remove AM/PM
    datetime = re.sub(r"[ ]?[APap][mM]", "", datetime).strip()

    # Language-specific cleaning
    datetime = re.sub(r"\b1er\b", "1", datetime)  # French
    datetime = re.sub(r"(\d+)η", r"\1", datetime)  # Greek
    datetime = re.sub(r"\bτου\b", "", datetime)  # Greek
    datetime = re.sub(r"°", "", datetime)  # Italian, Spanish
    datetime = re.sub(r"\bроку\b", "", datetime)  # Ukrainian

    # Remove extra spaces
    datetime = re.sub(r"\s+", " ", datetime).strip()

    return datetime



import re
import warnings

import pytest
from transformers import logging

from anonipy.definitions import Entity
from anonipy.anonymize.generators import (
    LLMLabelGenerator,
    MaskLabelGenerator,
    DateGenerator,
    NumberGenerator,
)

# disable transformers logging
logging.set_verbosity_error()

# =====================================
# Test Cases
# =====================================

DATETIME_STRS = [
    "2023-06-17 14:30:00",
    "17-06-2023 14:30:00",
    "06-17-2023 14:30:00",
    "2023/06/17 14:30:00",
    "17/06/2023 14:30:00",
    "06/17/2023 14:30:00",
    "2023.06.17 14:30:00",
    "17.06.2023 14:30:00",
    "06.17.2023 14:30:00",
    "2023 06 17 14:30:00",
    "17 06 2023 14:30:00",
    "06 17 2023 14:30:00",
    "2023-06-17 14:30",
    "17-06-2023 14:30",
    "06-17-2023 14:30",
    "2023/06/17 14:30",
    "17/06/2023 14:30",
    "06/17/2023 14:30",
    "2023.06.17 14:30",
    "17.06.2023 14:30",
    "06.17.2023 14:30",
    "2023 06 17 14:30",
    "17 06 2023 14:30",
    "06 17 2023 14:30",
    "2023-06-17 02:30 PM",
    "17-06-2023 02:30 PM",
    "06-17-2023 02:30 PM",
    "2023/06/17 02:30 PM",
    "17/06/2023 02:30 PM",
    "06/17/2023 02:30 PM",
    "2023.06.17 02:30 PM",
    "17.06.2023 02:30 PM",
    "06.17.2023 02:30 PM",
    "2023 06 17 02:30 PM",
    "17 06 2023 02:30 PM",
    "06 17 2023 02:30 PM",
    "June 17, 2023 14:30:00",
    "17 June 2023 14:30:00",
    "Jun 17, 2023 14:30:00",
    "17 Jun 2023 14:30:00",
    "June 17, 2023 02:30 PM",
    "17 June 2023 02:30 PM",
    "Jun 17, 2023 02:30 PM",
    "17 Jun 2023 02:30 PM",
    "Saturday, 17 June 2023 14:30:00",
    "Saturday, June 17, 2023 14:30:00",
    "Saturday, 17 June 2023 02:30 PM",
    "Saturday, June 17, 2023 02:30 PM",
    "2023-06-17",
    "17-06-2023",
    "06-17-2023",
    "2023/06/17",
    "17/06/2023",
    "06/17/2023",
    "2023.06.17",
    "17.06.2023",
    "06.17.2023",
    "2023 06 17",
    "17 06 2023",
    "06 17 2023",
    "June 17, 2023",
    "17 June 2023",
    "Jun 17, 2023",
    "17 Jun 2023",
    "Saturday, 17 June 2023",
    "Saturday, June 17, 2023",
]

# =====================================
# Helper functions
# =====================================


TEST_ORIGINAL_TEXT = """\
Medical Record

Patient Name: John Doe
Date of Birth: 15-01-1985
Date of Examination: 20-05-2024
Social Security Number: 123-45-6789

Examination Procedure:
John Doe underwent a routine physical examination. The procedure included measuring vital signs (blood pressure, heart rate, temperature), a comprehensive blood panel, and a cardiovascular stress test. The patient also reported occasional headaches and dizziness, prompting a neurological assessment and an MRI scan to rule out any underlying issues.

Medication Prescribed:

Ibuprofen 200 mg: Take one tablet every 6-8 hours as needed for headache and pain relief.
Lisinopril 10 mg: Take one tablet daily to manage high blood pressure.
Next Examination Date:
15-11-2024
"""

TEST_ENTITIES = {
    "name": Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        type="string",
    ),
    "name:pattern": Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        type="string",
        regex="Person: (.*)",
    ),
    "date": [
        Entity(
            text="20-05-2024",
            label="date",
            start_index=86,
            end_index=96,
            type="date",
        )
    ]
    + [
        Entity(
            text=str,
            label="date",
            start_index=86,
            end_index=86 + len(str),
            type="date",
        )
        for str in DATETIME_STRS
    ],
    "integer": Entity(
        text="123456789",
        label="integer",
        start_index=121,
        end_index=132,
        type="integer",
    ),
    "float": Entity(
        text="123,456,789.000",
        label="float",
        start_index=121,
        end_index=132,
        type="float",
    ),
    "custom": Entity(
        text="123-45-6789",
        label="custom",
        start_index=121,
        end_index=132,
        type="custom",
        regex="\\d{3}-\\d{2}-\\d{4}",
    ),
}


# =====================================
# Test LLM Label Generator
# =====================================


@pytest.fixture(scope="module")
def llm_label_generator():
    return LLMLabelGenerator()


def test_llm_label_generator_has_methods(llm_label_generator):
    assert hasattr(llm_label_generator, "generate")


def test_llm_label_generator_generate_default(llm_label_generator):
    entity = TEST_ENTITIES["name"]
    generated_text = llm_label_generator.generate(entity)
    regex = entity.get_regex_group() or entity.regex
    match = re.match(regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_llm_label_generator_generate_custom(llm_label_generator):
    entity = TEST_ENTITIES["name"]
    generated_text = llm_label_generator.generate(
        entity, add_entity_attrs="Spanish", temperature=0.5
    )
    regex = entity.get_regex_group() or entity.regex
    match = re.match(regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text

def test_llm_label_generator_generate_custom_user_prompt(llm_label_generator):
    entity = TEST_ENTITIES["name"]
    generated_text = llm_label_generator.generate(
        entity, add_entity_attrs="Spanish", temperature=0.5, user_prompt="Respond with the text 'TEST' only."
    )
    assert "TEST" in generated_text

def test_llm_label_generator_generate_custom_system_prompt(llm_label_generator):
    entity = TEST_ENTITIES["name"]
    generated_text = llm_label_generator.generate(
        entity, add_entity_attrs="Spanish", temperature=0.5, system_prompt="You are a helpful AI assistant for replying 'TEST'."
    )
    regex = entity.get_regex_group() or entity.regex
    match = re.match(regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text

def test_llm_label_generator_generate_pattern(llm_label_generator):
    entity = TEST_ENTITIES["name:pattern"]
    generated_text = llm_label_generator.generate(entity)
    regex = entity.get_regex_group() or entity.regex
    match = re.match(regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


# =====================================
# Test Mask Label Generator
# =====================================


@pytest.fixture(scope="module")
def mask_label_generator():
    return MaskLabelGenerator()


@pytest.fixture(autouse=True)
def suppress_warnings():
    warnings.filterwarnings("ignore", category=ImportWarning)
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=FutureWarning)


def test_mask_label_generator_has_methods(mask_label_generator):
    assert hasattr(mask_label_generator, "generate")


def test_mask_label_generator_generate_default(mask_label_generator):
    entity = TEST_ENTITIES["name"]
    generated_text = mask_label_generator.generate(entity, text=TEST_ORIGINAL_TEXT)
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


# =====================================
# Test Date Generator
# =====================================


@pytest.fixture(scope="module")
def date_generator():
    return DateGenerator(lang="en")


def test_date_generator_has_methods(date_generator):
    assert hasattr(date_generator, "generate")


def test_date_generator_generate_default(date_generator):
    entity = TEST_ENTITIES["date"][0]
    generated_text = date_generator.generate(entity)
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_date_generator_generate_custom_date_format():
    entity = TEST_ENTITIES["date"][0]
    generator = DateGenerator(date_format="dd-MM-yyyy")
    generated_text = generator.generate(entity)
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_date_generator_generate_non_matching_date_format():
    entity = TEST_ENTITIES["date"][0]
    generator = DateGenerator(date_format="yyyy-MM-dd")
    custom_date = generator.generate(entity, sub_variant="FIRST_DAY_OF_THE_MONTH")
    assert custom_date == "2024-05-01"


def test_date_generator_generate_first_day_of_the_month(date_generator):
    entity = TEST_ENTITIES["date"][0]
    generated_text = date_generator.generate(
        entity, sub_variant="FIRST_DAY_OF_THE_MONTH"
    )
    assert generated_text == "01-05-2024"


def test_date_generator_generate_last_day_of_the_month(date_generator):
    entity = TEST_ENTITIES["date"][0]
    generated_text = date_generator.generate(
        entity, sub_variant="LAST_DAY_OF_THE_MONTH"
    )
    assert generated_text == "31-05-2024"


def test_date_generator_generate_middle_of_the_month(date_generator):
    entity = TEST_ENTITIES["date"][0]
    generated_text = date_generator.generate(entity, sub_variant="MIDDLE_OF_THE_MONTH")
    assert generated_text == "15-05-2024"


def test_date_generator_generate_middle_of_the_year(date_generator):
    entity = TEST_ENTITIES["date"][0]
    generated_text = date_generator.generate(entity, sub_variant="MIDDLE_OF_THE_YEAR")
    assert generated_text == "01-07-2024"


def test_date_generator_generate_random(date_generator):
    entity = TEST_ENTITIES["date"][0]
    generated_text = date_generator.generate(entity, sub_variant="RANDOM")
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_date_generator_generate_uncorrect_type(date_generator):
    entity = TEST_ENTITIES["name"]
    with pytest.raises(ValueError):
        date_generator.generate(entity)


def test_date_generator_process_different_formats(date_generator):
    for entity in TEST_ENTITIES["date"]:
        try:
            date_generator.generate(entity, sub_variant="RANDOM")
        except ValueError:
            pytest.fail(
                f"date_generator.generate() raised ValueError unexpectedly for date: {entity.text}"
            )


# =====================================
# Test Number Generator
# =====================================


@pytest.fixture(scope="module")
def number_generator():
    return NumberGenerator()


def test_number_generator_has_methods(number_generator):
    assert hasattr(number_generator, "generate")


def test_number_generator_generate_integer(number_generator):
    entity = TEST_ENTITIES["integer"]
    generated_text = number_generator.generate(entity)
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_number_generator_generate_float(number_generator):
    entity = TEST_ENTITIES["float"]
    generated_text = number_generator.generate(entity)
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_number_generator_generate_custom(number_generator):
    entity = TEST_ENTITIES["custom"]
    generated_text = number_generator.generate(entity)
    match = re.match(entity.regex, generated_text)
    assert match is not None
    assert match.group(0) == generated_text


def test_number_generator_generate_uncorrect_type(number_generator):
    entity = TEST_ENTITIES["name"]
    with pytest.raises(ValueError):
        number_generator.generate(entity)



from anonipy.definitions import Entity

# =====================================
# Test Entity
# =====================================


def test_init_default():
    entity = Entity(
        text="test",
        label="test",
        start_index=0,
        end_index=4,
    )
    assert entity.text == "test"
    assert entity.label == "test"
    assert entity.start_index == 0
    assert entity.end_index == 4
    assert entity.score == 1.0
    assert entity.type is None
    assert entity.regex == ".*"


def test_init_custom():
    entity = Entity(
        text="test",
        label="test",
        start_index=0,
        end_index=4,
        score=0.89,
        type="test",
        regex="test",
    )
    assert entity.text == "test"
    assert entity.label == "test"
    assert entity.start_index == 0
    assert entity.end_index == 4
    assert entity.score == 0.89
    assert entity.type == "test"
    assert entity.regex == "test"



import random
import warnings
import datetime
from typing import Union

import dateparser
from babel.dates import format_datetime

from ...definitions import Entity
from ...utils.datetime_format import detect_datetime_format
from ...constants import DATE_TRANSFORM_VARIANTS, LANGUAGES

from .interface import GeneratorInterface

# =====================================
# Operation functions
# =====================================


def first_day_of_month(day: datetime.datetime, *args, **kwargs) -> datetime.datetime:
    """Returns the first day of the month of the given date.

    Args:
        day: The date to get the first day of the month from.

    Returns:
        The first day of the month of the given date.

    """
    return day.replace(day=1)


def last_day_of_month(day: datetime.datetime, *args, **kwargs) -> datetime.datetime:
    """Returns the last day of the month of the given date.

    Args:
        day: The date to get the last day of the month from.

    Returns:
        The last day of the month of the given date.

    """
    next_month = day.replace(day=28) + datetime.timedelta(days=4)
    return next_month - datetime.timedelta(days=next_month.day)


def middle_of_the_month(day: datetime.datetime, *args, **kwargs) -> datetime.datetime:
    """Returns the middle day of the month of the given date.

    Args:
        day: The date to get the middle day of the month from.

    Returns:
        The middle day of the month of the given date.

    """

    return day.replace(day=15)


def middle_of_the_year(day: datetime.datetime, *args, **kwargs) -> datetime.datetime:
    """Returns the middle day of the year of the given date.

    Args:
        day: The date to get the middle day of the year from.

    Returns:
        The middle day of the year of the given date.

    """

    return day.replace(month=7, day=1)


def random_date(
    day: datetime.datetime, sigma: int = 30, *args, **kwargs
) -> datetime.datetime:
    """Returns a random date within the given date range.

    The function returns a date within the range [day - sigma, day + sigma].

    Args:
        day: The date to get the random date from.
        sigma: The range of the random date in days.

    Returns:
        The random date within the given date range.

    """
    delta = random.randint(-sigma, sigma)
    return day + datetime.timedelta(days=delta)


DATE_VARIANTS_MAPPING = {
    DATE_TRANSFORM_VARIANTS.FIRST_DAY_OF_THE_MONTH: first_day_of_month,
    DATE_TRANSFORM_VARIANTS.LAST_DAY_OF_THE_MONTH: last_day_of_month,
    DATE_TRANSFORM_VARIANTS.MIDDLE_OF_THE_MONTH: middle_of_the_month,
    DATE_TRANSFORM_VARIANTS.MIDDLE_OF_THE_YEAR: middle_of_the_year,
    DATE_TRANSFORM_VARIANTS.RANDOM: random_date,
}

# =====================================
# Main class
# =====================================


class DateGenerator(GeneratorInterface):
    """The class representing the date generator.

    Examples:
        >>> from anonipy.anonymize.generators import DateGenerator
        >>> generator = DateGenerator(lang="de")
        >>> generator.generate(entity)

    Attributes:
        lang (str, LANGUAGES): The language of the text.
        date_format (str): The date format in which the date should be generated.
        day_sigma (int): The range of the random date in days.

    Methods:
        generate(entity, output_gen):
            Generate the date substitute based on the input parameters.

    """

    def __init__(
        self,
        *args,
        lang: Union[str, LANGUAGES] = "en",
        date_format: str = "auto",
        day_sigma: int = 30,
        **kwargs,
    ):
        """Initializes the date generator.

        Examples:
            >>> from anonipy.anonymize.generators import DateGenerator
            >>> generator = DateGenerator()

        Args:
            lang: The language of the text.
            date_format: The date format in which the date should be generated. More on date formats [see here](https://www.contensis.com/help-and-docs/guides/querying-your-content/zenql-search/date-formats).
            day_sigma: The range of the random date in days.

        """

        super().__init__(*args, **kwargs)
        self.date_format = date_format
        self.day_sigma = day_sigma

        if isinstance(lang, str) and lang in LANGUAGES.supported_languages():
            self.lang = lang
        elif isinstance(lang, LANGUAGES):
            self.lang = lang[0]
        else:
            raise Exception(f"Unknown lang value: {lang}")

    def generate(
        self,
        entity: Entity,
        *args,
        sub_variant: DATE_TRANSFORM_VARIANTS = DATE_TRANSFORM_VARIANTS.RANDOM,
        **kwargs,
    ) -> str:
        """Generate the entity substitute based on the input parameters.

        Args:
            entity: The entity to generate the date substitute from.
            sub_variant: The substitute function variant to use.

        Returns:
            The generated date substitute.

        Raises:
            ValueError: If the entity type is not `date` or `custom`.

        """

        if entity.type in ["custom"]:
            warnings.warn(
                "The entity type is `custom`. Make sure the generator is returning appropriate values."
            )
        elif entity.type not in ["date"]:
            raise ValueError("The entity type must be `date` to generate dates.")

        if not DATE_TRANSFORM_VARIANTS.is_valid(sub_variant):
            raise ValueError(
                f"The sub_variant must be one of {', '.join(DATE_TRANSFORM_VARIANTS.values())} to generate dates."
            )

        # Suppress all DeprecationWarnings from dateparser's internal strptime usage
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)

            # detect the date format
            if self.date_format == "auto":
                entity_date, date_format = detect_datetime_format(
                    entity.text, self.lang
                )
            else:
                entity_date = dateparser.parse(entity.text, languages=[self.lang])
                date_format = self.date_format

        # validate the input values
        if entity_date is None:
            raise ValueError(f"Entity `{entity.text}` is not a valid date.")
        if date_format is None or isinstance(date_format, ValueError):
            raise ValueError(f"Entity `{entity.text}` is not a valid date.")

        # generate the date substitute
        generate_date = DATE_VARIANTS_MAPPING[sub_variant](entity_date, self.day_sigma)
        return format_datetime(generate_date, format=date_format, locale=self.lang)



import re
import importlib
from typing import List, Tuple, Optional, Callable

from spacy import displacy, util
from spacy.tokens import Doc, Span
from spacy.language import Language
from spacy.matcher import Matcher

from ..helpers import (
    convert_spacy_to_entity,
    detect_repeated_entities,
    get_doc_entity_spans,
    set_doc_entity_spans,
    create_spacy_entities,
)
from ...constants import LANGUAGES
from ...definitions import Entity
from ...utils.colors import get_label_color

from .interface import ExtractorInterface


# ===============================================
# Extractor class
# ===============================================


class PatternExtractor(ExtractorInterface):
    """The class representing the pattern extractor.

    Examples:
        >>> from anonipy.constants import LANGUAGES
        >>> from anonipy.anonymize.extractors import PatternExtractor
        >>> labels = [{"label": "PERSON", "type": "string", "regex": "([A-Z][a-z]+ [A-Z][a-z]+)"}]
        >>> extractor = PatternExtractor(labels, lang=LANGUAGES.ENGLISH)
        >>> extractor("John Doe is a 19 year old software engineer.", detect_repeats=False)
        Doc, [Entity]

    Attributes:
        labels (List[dict]): The list of labels and patterns to extract.
        lang (str): The language of the text to extract.
        pipeline (Language): The spacy pipeline for extracting entities.
        token_matchers (Matcher): The spacy token pattern matcher.
        global_matchers (function): The global pattern matcher.

    Methods:
        __call__(self, text):
            Extract the entities from the text.
        display(self, doc):
            Display the entities in the text.

    """

    def __init__(
        self,
        labels: List[dict],
        *args,
        lang: LANGUAGES = LANGUAGES.ENGLISH,
        spacy_style: str = "ent",
        **kwargs,
    ):
        """Initialize the pattern extractor.

        Examples:
            >>> from anonipy.constants import LANGUAGES
            >>> from anonipy.anonymize.extractors import PatternExtractor
            >>> labels = [{"label": "PERSON", "type": "string", "regex": "([A-Z][a-z]+ [A-Z][a-z]+)"}]
            >>> extractor = PatternExtractor(labels, lang=LANGUAGES.ENGLISH)
            PatternExtractor()

        Args:
            labels: The list of labels and patterns to extract.
            lang: The language of the text to extract.
            spacy_style: The style the entities should be stored in the spacy doc. Options: `ent` or `span`.

        """

        super().__init__(labels, *args, **kwargs)
        self.lang = lang
        self.labels = labels
        self.spacy_style = spacy_style
        self.pipeline = self._prepare_pipeline()
        self.token_matchers = self._prepare_token_matchers()
        self.global_matchers = self._prepare_global_matchers()

    def __call__(
        self, text: str, detect_repeats: bool = False, *args, **kwargs
    ) -> Tuple[Doc, List[Entity]]:
        """Extract the entities from the text.

        Examples:
            >>> extractor("John Doe is a 19 year old software engineer.", detect_repeats=False)
            Doc, [Entity]

        Args:
            text: The text to extract entities from.
            detect_repeats: Whether to check text again for repeated entities.

        Returns:
            The spacy document.
            The list of extracted entities.

        """

        doc = self.pipeline(text)
        self.token_matchers(doc) if self.token_matchers else None
        self.global_matchers(doc) if self.global_matchers else None
        anoni_entities, spacy_entities = self._prepare_entities(doc)

        if detect_repeats:
            anoni_entities = detect_repeated_entities(
                doc, anoni_entities, self.spacy_style
            )

        create_spacy_entities(doc, anoni_entities, self.spacy_style)

        return doc, anoni_entities

    def display(self, doc: Doc, page: bool = False, jupyter: bool = None) -> str:
        """Display the entities in the text.

        Examples:
            >>> doc, entities = extractor("John Doe is a 19 year old software engineer.")
            >>> extractor.display(doc)
            HTML

        Args:
            doc: The spacy doc to display.
            page: Whether to display the doc in a web browser.
            jupyter: Whether to display the doc in a jupyter notebook.

        Returns:
            The HTML representation of the document and the extracted entities.

        """

        options = {
            "colors": {l["label"]: get_label_color(l["label"]) for l in self.labels}
        }
        return displacy.render(
            doc, style=self.spacy_style, options=options, page=page, jupyter=jupyter
        )

    # ===========================================
    # Private methods
    # ===========================================

    def _prepare_pipeline(self) -> Language:
        """Prepare the spacy pipeline.

        Prepares the pipeline for processing the text in the corresponding
        provided language.

        Returns:
            The spacy text processing and extraction pipeline.

        """

        # load the appropriate parser for the language
        module_lang, class_lang = self.lang[0].lower(), self.lang[1].lower().title()
        language_module = importlib.import_module(f"spacy.lang.{module_lang}")
        language_class = getattr(language_module, class_lang)
        # initialize the language parser
        nlp = language_class()
        nlp.add_pipe("sentencizer")
        return nlp

    def _prepare_token_matchers(self) -> Optional[Matcher]:
        """Prepare the token pattern matchers.

        Prepares the token pattern matchers for the provided labels.

        Returns:
            The spacy matcher object or None if no relevant labels are provided.

        """

        relevant_labels = list(filter(lambda l: "pattern" in l, self.labels))
        if len(relevant_labels) == 0:
            return None

        matcher = Matcher(self.pipeline.vocab)
        for label in relevant_labels:
            if isinstance(label["pattern"], list):
                on_match = self._create_add_event_ent(label["label"])
                matcher.add(label["label"], label["pattern"], on_match=on_match)
        return matcher

    def _prepare_global_matchers(self) -> Optional[Callable]:
        """Prepares the global pattern matchers.

        Prepares the global pattern matchers for the provided labels.

        Returns:
            The function used to match the patterns or None if no relevant labels are provided.

        """

        relevant_labels = list(filter(lambda l: "regex" in l, self.labels))
        if len(relevant_labels) == 0:
            return None

        def global_matchers(doc: Doc) -> None:
            for label in relevant_labels:
                for match in re.finditer(label["regex"], doc.text):
                    # define the entity span
                    start, end = match.span(1)
                    entity = doc.char_span(start, end, label=label["label"])
                    if not entity:
                        continue
                    entity._.score = 1.0
                    entities = [convert_spacy_to_entity(entity)]
                    # add the entity to the previous entity list
                    create_spacy_entities(doc, entities, self.spacy_style)

        return global_matchers

    def _prepare_entities(self, doc: Doc) -> Tuple[List[Entity], List[Span]]:
        """Prepares the anonipy and spacy entities.

        Args:
            doc: The spacy doc to prepare.

        Returns:
            The list of anonipy entities.
            The list of spacy entities.

        """

        # TODO: make this part more generic
        anoni_entities = []
        spacy_entities = []
        for e in get_doc_entity_spans(doc, self.spacy_style):
            label = list(filter(lambda x: x["label"] == e.label_, self.labels))[0]
            anoni_entities.append(convert_spacy_to_entity(e, **label))
            spacy_entities.append(e)
        return anoni_entities, spacy_entities

    def _create_add_event_ent(self, label: str) -> Callable:
        """Create the add event entity function

        Args:
            label: The identified label entity.

        Returns:
            The function used to add the entity to the spacy doc.

        """

        def add_event_ent(matcher, doc, i, matches):
            # define the entity span
            _, start, end = matches[i]
            entity = Span(doc, start, end, label=label)
            if not entity:
                return
            entity._.score = 1.0
            entities = [convert_spacy_to_entity(entity)]
            create_spacy_entities(doc, entities, self.spacy_style)

        return add_event_ent



"""Module containing the anonymization modules and utility.

The `anonymize` module provides a set of anonymization modules and utility,
including `extractors`, `generators`, and `strategies`. In addition, it provides
methods for anonymizing text based on a list of replacements, as well as a
`pipeline` class for automating the anonymization process.

Modules:
    extractors: The module containing the extractor classes.
    generators: The module containing the generator classes.
    strategies: The module containing the strategy classes.
    pipeline: The module containing the pipeline class.

Methods:
    anonymize(text, replacements):
        Anonymize the text based on the replacements.

"""

from . import extractors
from . import generators
from . import strategies
from . import pipeline
from .helpers import anonymize


__all__ = ["extractors", "generators", "strategies", "anonymize", "pipeline"]



from ...definitions import Entity


# =====================================
# Main functions
# =====================================


class GeneratorInterface:
    """The class representing the generator interface.

    All generators should inherit from this class.

    Methods:
        generate(entity):
            Generate a substitute for the entity.

    """

    def __init__(self, *args, **kwargs):
        pass

    def generate(self, entity: Entity, *args, **kwargs):
        pass



import warnings
from typing import Tuple, List

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig

from ...utils.package import is_installed_with
from .interface import GeneratorInterface
from ...definitions import Entity

# =====================================
# Main class
# =====================================


class LLMLabelGenerator(GeneratorInterface):
    """The class representing the LLM label generator.

    Examples:
        >>> from anonipy.anonymize.generators import LLMLabelGenerator
        >>> generator = LLMLabelGenerator()
        >>> generator.generate(entity)

    Attributes:
        model (models.Transformers): The model used to generate the label substitutes.

    Methods:
        generate(entity, entity_prefix, temperature):
            Generate the label based on the entity.

    """

    def __init__(
        self,
        *args,
        model_name: str = "HuggingFaceTB/SmolLM2-1.7B-Instruct",
        use_gpu: bool = False,
        use_quant: bool = False,
        **kwargs,
    ):
        """Initializes the LLM label generator.

        Args:
            model_name: The name of the model to use.
            use_gpu: Whether to use GPU or not.
            use_quant: Whether to use quantization or not.

        Examples:
            >>> from anonipy.anonymize.generators import LLMLabelGenerator
            >>> generator = LLMLabelGenerator()
            LLMLabelGenerator()

        """

        super().__init__(*args, **kwargs)

        if use_gpu and not torch.cuda.is_available():
            warnings.warn(
                "The use_gpu=True flag requires GPU/CUDA, but it is not available. Setting use_gpu=False."
            )
            use_gpu = False

        if use_quant and not is_installed_with(["quant", "all"]):
            warnings.warn(
                "The use_quant=True flag requires the 'quant' extra dependencies, but they are not installed. Setting use_quant=False."
            )
            use_quant = False

        self.model, self.tokenizer = self._prepare_model_and_tokenizer(
            model_name, use_gpu, use_quant
        )

    def generate(
        self,
        entity: Entity,
        *args,
        add_entity_attrs: str = "",
        temperature: float = 1.0,
        top_p: float = 0.95,
        system_prompt: str = "You are a helpful AI assistant for generating replacements for text entities.",
        user_prompt: str = "What is a random {add_entity_attrs} {entity.label} replacement for {entity.text}? Respond only with the replacement.",
        **kwargs,
    ) -> str:
        """Generate the substitute for the entity based on it's attributes.

        Examples:
            >>> from anonipy.anonymize.generators import LLMLabelGenerator
            >>> generator = LLMLabelGenerator()
            >>> generator.generate(entity)
            label

        Args:
            entity: The entity to generate the label from.
            add_entity_attrs (str, optional): Additional entity attribute description to add to the generation.
            temperature: The temperature to use for the generation.
            top_p: The top p to use for the generation.
            system_prompt (str, optional): The system prompt to use for the generation.
            user_prompt (str, optional): The user prompt to use for the generation.

        Returns:
            The generated entity label substitute.

        """

        message = [
            {
                "role": "system",
                "content": system_prompt,
            },
            {
                "role": "user",
                "content": user_prompt.format(
                    add_entity_attrs=add_entity_attrs, entity=entity
                ),
            },
        ]
        return self._generate_response(message, temperature, top_p)

    # =================================
    # Private methods
    # =================================

    def _prepare_model_and_tokenizer(
        self, model_name: str, use_gpu: bool, use_quant: bool
    ) -> Tuple[AutoModelForCausalLM, AutoTokenizer]:
        """Prepares the model and tokenizer.

        Args:
            model_name: The name of the model to use.

        Returns:
            The huggingface model.
            The huggingface tokenizer.

        """

        # prepare the model
        device = torch.device(
            "cuda" if use_gpu and torch.cuda.is_available() else "cpu"
        )
        dtype = torch.float32 if device.type == "cpu" else torch.float16

        model = self._load_model(model_name, device, dtype, use_quant, use_gpu)
        tokenizer = self._load_tokenizer(model_name)

        return model, tokenizer

    def _load_model(
        self,
        model_name: str,
        device: torch.device,
        dtype: torch.dtype,
        use_quant: bool,
        use_gpu: bool,
    ) -> AutoModelForCausalLM:
        """Load the model with appropriate configuration.

        Args:
            model_name: The name of the model to use.
            device: The device to use for the model.
            dtype: The data type to use for the model.
            use_quant: Whether to use quantization or not.
            use_gpu: Whether to use GPU or not.

        Returns:
            The huggingface model.

        """
        if use_quant and use_gpu:
            quant_config = BitsAndBytesConfig(
                load_in_8bit=True, bnb_4bit_compute_dtype=dtype
            )
            return AutoModelForCausalLM.from_pretrained(
                model_name,
                device_map=device,
                torch_dtype=dtype,
                quantization_config=quant_config,
            )

        if use_quant:
            warnings.warn(
                "Quantization is only supported on GPU, but use_gpu=False. Loading model without quantization."
            )

        return AutoModelForCausalLM.from_pretrained(
            model_name, device_map=device, torch_dtype=dtype
        )

    def _load_tokenizer(self, model_name: str) -> AutoTokenizer:
        """Load the tokenizer with appropriate configuration.

        Args:
            model_name: The name of the model to use.

        Returns:
            The huggingface tokenizer.
        """
        return AutoTokenizer.from_pretrained(
            model_name, padding_side="right", use_fast=False
        )

    def _generate_response(
        self, message: List[dict], temperature: float, top_p: float
    ) -> str:
        """Generate the response from the LLM.

        Args:
            message: The message to generate the response from.
            temperature: The temperature to use for the generation.
            top_p: The top p to use for the generation.

        Returns:
            The generated response.

        """

        # tokenize the message
        input_ids = self.tokenizer.apply_chat_template(
            message, tokenize=True, return_tensors="pt", add_generation_prompt=True
        ).to(self.model.device)

        # create attention mask (1 for all tokens)
        attention_mask = torch.ones_like(input_ids)

        # set pad token id if not set
        if self.tokenizer.pad_token_id is None:
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=".*the `logits` model output.*")

            # generate the response
            with torch.no_grad():
                output_ids = self.model.generate(
                    input_ids,
                    attention_mask=attention_mask,
                    max_new_tokens=50,
                    temperature=temperature,
                    top_p=top_p,
                    do_sample=True,
                    pad_token_id=self.tokenizer.pad_token_id,
                )

        # decode the response
        response = self.tokenizer.decode(
            output_ids[0][len(input_ids[0]) :], skip_special_tokens=True
        )
        return response



"""The module containing the `language_detector` utilities.

The `language_detector` module contains the `LanguageDetector` class, which is
used to detect the language of a text.

Classes:
    LanguageDetector: The class representing the language detector.

"""

from typing import Tuple

from lingua import LanguageDetectorBuilder

# =====================================
# Main class
# =====================================


class LanguageDetector:
    """The class representing the language detector.

    Examples:
        >>> from anonipy.utils.language_detector import LanguageDetector
        >>> detector = LanguageDetector()
        >>> detector.detect(text)

    Attributes:
        detector (lingua.LanguageDetector): The language detector.

    Methods:
        __call__(text, output_standard):
            Detect the language of a text. Calls the `detect` method.

        detect(text, output_standard):
            Detect the language of a text.

    """

    def __init__(self, low_accuracy: bool = False):
        """Initializes the language detector.

        Examples:
            >>> from anonipy.utils.language_detector import LanguageDetector
            >>> detector = LanguageDetector()

        Args:
            low_accuracy: Whether to use the low accuracy mode.

        """

        # Prepare the language detector for all languages
        builder = LanguageDetectorBuilder.from_all_languages()
        builder = (
            builder.with_low_accuracy_mode()
            if low_accuracy
            else builder.with_preloaded_language_models()
        )
        self.detector = builder.build()

    def __call__(
        self, text: str, output_standard: str = "iso_code_639_1"
    ) -> Tuple[str, str]:
        """Detects the language of a text by calling the `detect` method.

        Examples:
            >>> from anonipy.utils.language_detector import LanguageDetector
            >>> detector = LanguageDetector()
            >>> detector(text)

        Args:
            text: The text to detect the language of.
            output_standard: The output standard.

        Returns:
            The language code.
            The full name of the language.

        """

        return self.detect(text, output_standard)

    def detect(
        self, text: str, output_standard: str = "iso_code_639_1"
    ) -> Tuple[str, str]:
        """Detects the language of a text.

        Examples:
            >>> from anonipy.utils.language_detector import LanguageDetector
            >>> detector = LanguageDetector()
            >>> detector.detect(text)

        Args:
            text: The text to detect the language of.
            output_standard: The output standard.

        Returns:
            The language code.
            The full name of the language.

        """

        language = self.detector.detect_language_of(text)
        iso_code = getattr(language, output_standard).name.lower()
        full_name = language.name.lower().title()
        # use the correct name for Slovenian
        full_name = "Slovenian" if full_name == "Slovene" else full_name
        return iso_code, full_name



import re
from typing import List, Union, Tuple, Iterable, Set
import itertools

from spacy import util
from spacy.tokens import Span, Doc

from ..definitions import Entity, Replacement
from ..constants import ENTITY_TYPES


# =====================================
# Entity converters
# =====================================


def convert_spacy_to_entity(
    entity: Span,
    type: ENTITY_TYPES = None,
    regex: Union[str, re.Pattern] = None,
    *args,
    **kwargs,
) -> Entity:
    """Convert a spacy entity to an anonipy entity object.

    Args:
        entity: The spacy Span representing the entity to convert.
        type: The type of the entity.
        regex: The regular expression the entity must match.

    Returns:
        The converted anonipy entity object.

    """

    return Entity(
        entity.text,
        entity.label_,
        entity.start_char,
        entity.end_char,
        entity._.score,
        type,
        regex,
    )


# =====================================
# Anonymization function
# =====================================


def anonymize(text: str, replacements: List[Replacement]) -> str:
    """Anonymize a text based on a list of replacements.

    Examples:
        >>> from anonipy.anonymize import anonymize
        >>> anonymize(text, replacements)

    Args:
        text: The text to anonymize.
        replacements: The list of replacements to apply.

    Returns:
        The anonymized text.

    """

    s_replacements = sorted(replacements, key=lambda x: x["start_index"], reverse=True)

    anonymized_text = text
    for replacement in s_replacements:
        anonymized_text = (
            anonymized_text[: replacement["start_index"]]
            + replacement["anonymized_text"]
            + anonymized_text[replacement["end_index"] :]
        )
    return anonymized_text, s_replacements[::-1]


# =====================================
# Entity helpers
# =====================================


def merge_entities(extractor_outputs: List[Tuple[Doc, List[Entity]]]) -> List[Entity]:
    """Merges the entities returned by the extractors.

    Args:
        extractor_outputs: The list of extractor outputs.

    Returns:
        The merged entities list.

    """

    if len(extractor_outputs) == 0:
        return []
    if len(extractor_outputs) == 1:
        return extractor_outputs[0][1]

    joint_entities = filter_entities(
        list(itertools.chain.from_iterable([entity[1] for entity in extractor_outputs]))
    )
    return joint_entities


def filter_entities(entities: Iterable[Entity]) -> List[Entity]:
    """Filters the entities based on their start and end indices.

    Args:
        entities: The entities to filter.

    Returns:
        The filtered entities.

    """

    def get_sort_key(entity):
        return (
            entity.end_index - entity.start_index,
            -entity.start_index,
        )

    sorted_entities = sorted(entities, key=get_sort_key, reverse=True)
    result = []
    seen_tokens: Set[int] = set()
    for entity in sorted_entities:
        # Check for end - 1 here because boundaries are inclusive
        if (
            entity.start_index not in seen_tokens
            and entity.end_index - 1 not in seen_tokens
        ):
            result.append(entity)
            seen_tokens.update(range(entity.start_index, entity.end_index))
    result = sorted(result, key=lambda entity: entity.start_index)
    return result


def detect_repeated_entities(
    doc: Doc, entities: List[Entity], spacy_style: str
) -> List[Entity]:
    """Detects repeated entities in the text.

    Args:
        doc: The spacy doc to detect entities in.
        entities: The entities to detect.
        spacy_style: The style the entities should be stored in the spacy doc.

    Returns:
        The list of all entities.

    """

    repeated_entities = []

    for entity in entities:
        matches = re.finditer(re.escape(entity.text), doc.text)
        for match in matches:
            start_index, end_index = match.start(), match.end()
            if start_index == entity.start_index and end_index == entity.end_index:
                continue
            repeated_entities.append(
                Entity(
                    text=entity.text,
                    label=entity.label,
                    start_index=start_index,
                    end_index=end_index,
                    score=entity.score,
                    type=entity.type,
                    regex=entity.regex,
                )
            )

    filtered_entities = filter_entities(entities + repeated_entities)
    final_entities = sorted(filtered_entities, key=lambda e: e.start_index)

    return final_entities


# ====================================
# Spacy helpers
# ====================================


def create_spacy_entities(doc: Doc, entities: List[Entity], spacy_style: str) -> None:
    """Create spacy entities in the spacy doc.

    Args:
        doc: The spacy doc to create entities in.
        entities: The entities to create.
        spacy_style: The style the entities should be stored in the spacy doc.

    """

    updated_spans = get_doc_entity_spans(doc, spacy_style)

    for entity in entities:
        span = doc.char_span(entity.start_index, entity.end_index, label=entity.label)
        if not span or span in updated_spans:
            continue
        span._.score = entity.score
        if spacy_style == "ent":
            updated_spans = util.filter_spans(updated_spans + (span,))
        elif spacy_style == "span":
            updated_spans.append(span)
        else:
            raise ValueError(f"Invalid spacy style: {spacy_style}")

    set_doc_entity_spans(doc, updated_spans, spacy_style)


def get_doc_entity_spans(doc: Doc, spacy_style: str) -> List[Span]:
    """Get the spacy doc entity spans.

    Args:
        doc: The spacy doc to get the entity spans from.
        spacy_style: The spacy style to use.

    Returns:
        The list of entity spans.

    """

    if spacy_style == "ent":
        return doc.ents
    if spacy_style == "span":
        if "sc" not in doc.spans:
            doc.spans["sc"] = []
        return doc.spans["sc"]
    raise ValueError(f"Invalid spacy style: {spacy_style}")


def set_doc_entity_spans(doc: Doc, entities: List[Span], spacy_style: str) -> None:
    """Set the spacy doc entity spans.

    Args:
        doc: The spacy doc to set the entity spans.
        entities: The entity spans to assign the doc.
        spacy_style: The spacy style to use.

    """

    if spacy_style == "ent":
        doc.ents = entities
    elif spacy_style == "span":
        doc.spans["sc"] = entities
    else:
        raise ValueError(f"Invalid spacy style: {spacy_style}")



from typing import List
from ...definitions import Entity

# =====================================
# Main class
# =====================================


class StrategyInterface:
    """The class representing the strategy interface.

    All strategies should inherit from this class.

    Methods:
        anonymize(text, entities):
            Anonymize the text based on the entities.

    """

    def __init__(self, *args, **kwargs):
        pass

    def anonymize(self, text: str, entities: List[Entity], *args, **kwargs):
        pass



import re
from typing import List, Tuple

from .interface import StrategyInterface
from ...definitions import Entity, Replacement
from ..helpers import anonymize

# =====================================
# Main class
# =====================================


class MaskingStrategy(StrategyInterface):
    """The class representing the masking strategy.

    Examples:
        >>> from anonipy.anonymize.strategies import MaskingStrategy
        >>> strategy = MaskingStrategy()
        >>> strategy.anonymize(text, entities)

    Attributes:
        substitute_label (str): The label to substitute in the anonymized text.

    Methods:
        anonymize(text, entities):
            Anonymize the text based on the entities.

    """

    def __init__(self, substitute_label: str = "*", *args, **kwargs):
        """Initializes the masking strategy.

        Examples:
            >>> from anonipy.anonymize.strategies import MaskingStrategy
            >>> strategy = MaskingStrategy()

        Args:
            substitute_label: The label to substitute in the anonymized text.

        """

        super().__init__(*args, **kwargs)
        self.substitute_label = substitute_label or "*"

    def anonymize(
        self, text: str, entities: List[Entity], *args, **kwargs
    ) -> Tuple[str, List[Replacement]]:
        """Anonymize the text using the masking strategy.

        Examples:
            >>> from anonipy.anonymize.strategies import MaskingStrategy
            >>> strategy = MaskingStrategy()
            >>> strategy.anonymize(text, entities)

        Args:
            text: The text to anonymize.
            entities: The list of entities to anonymize.

        Returns:
            The anonymized text.
            The list of applied replacements.

        """

        replacements = [self._create_replacement(ent) for ent in entities]
        anonymized_text, replacements = anonymize(text, replacements)
        return anonymized_text, replacements

    # ===========================================
    # Private methods
    # ===========================================

    def _create_replacement(self, entity: Entity) -> Replacement:
        """Creates a replacement for the entity.

        Args:
            entity: The entity to create the replacement for.

        Returns:
            The created replacement.

        """

        mask = self._create_mask(entity)
        return {
            "original_text": entity.text,
            "label": entity.label,
            "start_index": entity.start_index,
            "end_index": entity.end_index,
            "anonymized_text": mask,
        }

    def _create_mask(self, entity: Entity) -> str:
        """Creates a mask for the entity.

        Args:
            entity: The entity to create the mask for.

        Returns:
            The created mask.

        """

        # TODO: add random length substitution
        return " ".join(
            [
                self.substitute_label * len(chunk)
                for chunk in re.split(r"\s+", entity.text)
            ]
        )



"""Module containing the `regex` related utilities.

The `regex` module provides the regex definitions and functions used within the package.

Classes:
    RegexMapping: The class representing the mapping for data type to the corresponding regex.

Attributes:
    REGEX_STRING (str): The regex definition for string.
    REGEX_INTEGER (str): The regex definition for integer.
    REGEX_FLOAT (str): The regex definition for float.
    REGEX_DATE (str): The regex definition for date.
    REGEX_EMAIL_ADDRESS (str): The regex definition for email address.
    REGEX_PHONE_NUMBER (str): The regex definition for phone number.
    REGEX_WEBSITE_URL (str): The regex definition for website URL.

"""

from collections import defaultdict

from ..constants import ENTITY_TYPES


# =====================================
# Regex definitions
# =====================================

REGEX_STRING = r".*"
"""The regex definition for string."""

REGEX_INTEGER = r"\d+"
"""The regex definition for integer."""

REGEX_FLOAT = r"[\d\.,]+"
"""The regex definition for float."""

REGEX_DATE = (
    r"("
    r"(\d{4}[-/.\s]\d{1,2}[-/.\s]\d{1,2}(?:[ T]\d{2}:\d{2}:\d{2})?)|"
    r"(\d{1,2}[-/.\s]\d{1,2}[-/.\s]\d{4}(?:[ T]\d{2}:\d{2}:\d{2})?)|"
    r"(\d{1,2}[-/.\s]\d{1,2}[-/.\s]\d{4}(?:[ T]\d{2}:\d{2})?)|"
    r"(\d{4}[-/.\s]\d{1,2}[-/.\s]\d{1,2}(?:[ T]\d{2}:\d{2})?)|"
    r"(\d{4}[-/.\s]\d{1,2}[-/.\s]\d{1,2}(?:[ T]\d{2}:\d{2} [APap][mM])?)|"
    r"(\d{1,2}[-/.\s]\d{1,2}[-/.\s]\d{4}(?:[ T]\d{2}:\d{2} [APap][mM])?)|"
    r"(\d{1,2}[-/.\s]\d{1,2}[-/.\s]\d{4}(?:[ ]?\d{2}:\d{2}:\d{2})?)|"
    r"(\d{4}[-/.\s]\d{1,2}[-/.\s]\d{1,2}(?:[ ]?\d{2}:\d{2}:\d{2})?)|"
    # English dates
    r"(\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{4}(?:[ ]?\d{2}:\d{2}:\d{2})?)|"
    r"(\d{1,2}[ ](Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[ ]\d{4}(?:[ ]?\d{2}:\d{2}:\d{2})?)|"
    r"(\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{4}(?:[ ]?\d{2}:\d{2}[ ]?[APap][mM])?)|"
    r"(\d{1,2}[ ](Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[ ]\d{4}(?:[ ]?\d{2}:\d{2}[ ]?[APap][mM])?)|"
    r"([A-Za-z]+,[ ]\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December),?[ ]\d{4}(?:[ ]?\d{2}:\d{2}:\d{2})?)|"
    r"([A-Za-z]+,[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{1,2},?[ ]\d{4}(?:[ ]?\d{2}:\d{2}:\d{2})?)|"
    r"([A-Za-z]+,[ ]\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December),?[ ]\d{4}(?:[ ]?\d{2}:\d{2}[ ]?[APap][mM])?)|"
    r"([A-Za-z]+,[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{1,2},?[ ]\d{4}(?:[ ]?\d{2}:\d{2}[ ]?[APap][mM])?)|"
    # Dutch dates
    r"(\d{1,2}[\.]?[ ](januari|februari|maart|april|mei|juni|juli|augustus|september|oktober|november|december)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](jan|feb|mrt|apr|mei|jun|jul|aug|sep|okt|nov|dec)[\.]?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ]\d{1,2}[\.]?[ ](januari|februari|maart|april|mei|juni|juli|augustus|september|oktober|november|december),?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ](januari|februari|maart|april|mei|juni|juli|augustus|september|oktober|november|december)[ ]\d{1,2},?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # French dates
    r"(\d{1,2}(er)?[ ](janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}(er)?[ ](jan|févr|mars|avr|mai|juin|juil|août|sept|oct|nov|déc)[\.]?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zÀ-ÿ]+,?[ ]\d{1,2}(er)?[ ](janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre),?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zÀ-ÿ]+,?[ ](janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre)[ ]\d{1,2}(er)?,?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # German dates
    r"(\d{1,2}[\.]?[ ](Januar|Februar|März|April|Mai|Juni|Juli|August|September|Oktober|November|Dezember)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](Jan|Feb|Mär|Apr|Mai|Jun|Jul|Aug|Sep|Okt|Nov|Dez)[\.]?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zäöüßÄÖÜ]+,?[ ]\d{1,2}[\.]?[ ](Januar|Februar|März|April|Mai|Juni|Juli|August|September|Oktober|November|Dezember)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zäöüßÄÖÜ]+,?[ ](Januar|Februar|März|April|Mai|Juni|Juli|August|September|Oktober|November|Dezember)[ ]\d{1,2}[\.]?,?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # Greek dates
    r"(\d{1,2}η?[ ](Ιανουαρίου|Φεβρουαρίου|Μαρτίου|Απριλίου|Μαΐου|Ιουνίου|Ιουλίου|Αυγούστου|Σεπτεμβρίου|Οκτωβρίου|Νοεμβρίου|Δεκεμβρίου)( του)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}i?[ ](Ianouariou|Fevrouariou|Martiou|Apriliou|Maiou|Iouniou|Iouliou|Avgoustou|Septemvriou|Oktovriou|Noemvriou|Dekemvriou)( tou)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}η?[ ](Ιαν|Φεβ|Μάρ|Απρ|Μάι|Ιούν|Ιούλ|Aυγ|Σεπ|Οκτ|Νοε|Δεκ)[\.]?( του)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}i?[ ](Ian|Feb|Mar|Apr|Mai|Iou|Ioul|Avg|Sep|Okt|Noe|Dek)[\.]?( tou)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([\u0370-\u03FF]+,?[ ]\d{1,2}η?[ ](Ιανουαρίου|Φεβρουαρίου|Μαρτίου|Απριλίου|Μαΐου|Ιουνίου|Ιουλίου|Αυγούστου|Σεπτεμβρίου|Οκτωβρίου|Νοεμβρίου|Δεκεμβρίου),?( του)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ]\d{1,2}i?[ ](Ianouariou|Fevrouariou|Martiou|Apriliou|Maiou|Iouniou|Iouliou|Avgoustou|Septemvriou|Oktovriou|Noemvriou|Dekemvriou),?( tou)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([\u0370-\u03FF]+,?[ ](Ιανουαρίου|Φεβρουαρίου|Μαρτίου|Απριλίου|Μαΐου|Ιουνίου|Ιουλίου|Αυγούστου|Σεπτεμβρίου|Οκτωβρίου|Νοεμβρίου|Δεκεμβρίου)[ ]\d{1,2}η?,?( του)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ](Ianouariou|Fevrouariou|Martiou|Apriliou|Maiou|Iouniou|Iouliou|Avgoustou|Septemvriou|Oktovriou|Noemvriou|Dekemvriou)[ ]\d{1,2}i?,?( tou)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # Italian dates
    r"(\d{1,2}°?[ ](gennaio|febbraio|marzo|aprile|maggio|giugno|luglio|agosto|settembre|ottobre|novembre|dicembre)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}°?[ ](gen|feb|mar|apr|mag|giu|lug|ago|set|ott|nov|dic)[\.]?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zÀ-ÿ]+,?[ ]\d{1,2}°?[ ](gennaio|febbraio|marzo|aprile|maggio|giugno|luglio|agosto|settembre|ottobre|novembre|dicembre),?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zÀ-ÿ]+,?[ ](gennaio|febbraio|marzo|aprile|maggio|giugno|luglio|agosto|settembre|ottobre|novembre|dicembre)[ ]\d{1,2}°?,?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # Slovenian dates
    r"(\d{1,2}[\.]?[ ](januar|februar|marec|april|maj|junij|julij|avgust|september|oktober|november|december)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](januarja|februarja|marca|aprila|maja|junija|julija|avgusta|septembra|oktobra|novembra|decembra)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](jan|feb|mar|apr|maj|jun|jul|avg|sep|okt|nov|dec)[\.]?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zčšžČŠŽ]+,?[ ]\d{1,2}[\.]?[ ](januar|februar|marec|april|maj|junij|julij|avgust|september|oktober|november|december)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zčšžČŠŽ]+,?[ ]\d{1,2}[\.]?[ ](januarja|februarja|marca|aprila|maja|junija|julija|avgusta|septembra|oktobra|novembra|decembra)[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # Spanish dates
    r"(\d{1,2}°?( de)?[ ](enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)( de)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}°?( de)?[ ](ene|feb|mar|abr|may|jun|jul|ago|sept|oct|nov|dic)[\.]?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zÀ-ÿ]+,?[ ]\d{1,2}°?( de)?[ ](enero|febrero|marzo|abril|mayo|junio|julio|augusto|septiembre|octubre|noviembre|diciembre),?( de)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-zÀ-ÿ]+,?[ ](enero|febrero|marzo|abril|mayo|junio|julio|augusto|septiembre|octubre|noviembre|diciembre)[ ]\d{1,2}°?,?( de)?[ ]\d{4}(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    # Ukrainian dates
    r"(\d{1,2}[\.]?[ ](січень|лютий|березень|квітень|травень|червень|липень|серпень|вересень|жовтень|листопад|грудень)[ ]\d{4}( року| рік)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](sichen|lyutyi|berezen|kviten|traven|cherven|lypen|serpen|veresen|zhovten|lystopad|gruden)[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](січ|лют|бер|кві|тра|чер|лип|сер|вер|жов|лис|гру)[\.]?[ ]\d{4}( року| рік)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](sich|lyut|ber|kvi|tra|cher|lyp|ser|ver|zhov|lys|gru)[\.]?[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([А-ЩЬЮЯҐЄІЇа-щьюяґєії]+,?[ ]\d{1,2}[\.]?[ ](січень|лютий|березень|квітень|травень|червень|липень|серпень|вересень|жовтень|листопад|грудень)[ ]\d{4}( року)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ]\d{1,2}[\.]?[ ](sichen|lyutyi|berezen|kviten|traven|cherven|lypen|serpen|veresen|zhovten|lystopad|gruden)[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([А-ЩЬЮЯҐЄІЇа-щьюяґєії]+,?[ ](січень|лютий|березень|квітень|травень|червень|липень|серпень|вересень|жовтень|листопад|грудень)[ ]\d{1,2}[\.]?,?[ ]\d{4}( року)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ](sichen|lyutyi|berezen|kviten|traven|cherven|lypen|serpen|veresen|zhovten|lystopad|gruden)[ ]\d{1,2}[\.]?,?[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](січня|лютого|березня|квітня|травня|червня|липня|серпня|вересня|жовтня|листопада|грудня)[ ]\d{4}( року| рік)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"(\d{1,2}[\.]?[ ](sichnia|liutoho|bereznia|kvitnia|travniia|chervnia|lypnia|serpnia|veresnia|zhovtnia|lystopada|hrudnia)[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([А-ЩЬЮЯҐЄІЇа-щьюяґєії]+,?[ ]\d{1,2}[\.]?[ ](січня|лютого|березня|квітня|травня|червня|липня|серпня|вересня|жовтня|листопада|грудня)[ ]\d{4}( року| рік)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ]\d{1,2}[\.]?[ ](sichnia|liutoho|bereznia|kvitnia|travniia|chervnia|lypnia|serpnia|veresnia|zhovtnia|lystopada|hrudnia)[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([А-ЩЬЮЯҐЄІЇа-щьюяґєії]+,?[ ](січня|лютого|березня|квітня|травня|червня|липня|серпня|вересня|жовтня|листопада|грудня)[ ]\d{1,2}[\.]?,?[ ]\d{4}( року| рік)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)|"
    r"([A-Za-z]+,?[ ](sichnia|liutoho|bereznia|kvitnia|travniia|chervnia|lypnia|serpnia|veresnia|zhovtnia|lystopada|hrudnia)[ ]\d{1,2}[\.]?,?[ ]\d{4}( roku)?(?:[ ]?\d{2}:\d{2}(?::\d{2})?)?)"
    r")"
)
"""The regex definition for dates."""

REGEX_EMAIL_ADDRESS = (
    r"[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*"
)
"""The regex definition for email addresses."""

REGEX_PHONE_NUMBER = (
    r"[(]?[\+]?[(]?[0-9]{1,3}[)]?[-\s\.]?([0-9]{2,}[-\s\.]?){2,}([0-9]{3,})"
)
"""The regex definition for phone numbers."""

REGEX_WEBSITE_URL = r"((https?|ftp|smtp):\/\/)?(www.)?([a-zA-Z0-9]+\.)+[a-z]{2,}(\/[a-zA-Z0-9#\?\_\.\=\-\&]+|\/?)*"
"""The regex definition for website URLs."""


# =====================================
# Define the regex definitions
# =====================================


class RegexMapping:
    """The class representing the regex mapping.

    Examples:
        >>> from anonipy.anonymize.regex import regex_mapping
        >>> regex_mapping["string"]
        ".*"

    Attributes:
        regex_mapping (defaultdict):
            The mapping between the data type and the corresponding regex.

    Methods:
        __getitem__(type):
            Gets the regex for the given type.

    """

    def __init__(self):
        """Initialize the regex mapping.

        Examples:
            >>> from anonipy.anonymize.regex import RegexMapping
            >>> regex_mapping = RegexMapping()

        """

        self.regex_mapping = defaultdict(lambda: ".*")
        # Define the regex mappings
        self.regex_mapping[ENTITY_TYPES.STRING] = REGEX_STRING
        self.regex_mapping[ENTITY_TYPES.INTEGER] = REGEX_INTEGER
        self.regex_mapping[ENTITY_TYPES.FLOAT] = REGEX_FLOAT
        self.regex_mapping[ENTITY_TYPES.DATE] = REGEX_DATE
        self.regex_mapping[ENTITY_TYPES.EMAIL] = REGEX_EMAIL_ADDRESS
        self.regex_mapping[ENTITY_TYPES.PHONE_NUMBER] = REGEX_PHONE_NUMBER
        self.regex_mapping[ENTITY_TYPES.WEBSITE_URL] = REGEX_WEBSITE_URL

    def __getitem__(self, regex_type: str) -> str:
        """Gets the regex for the given type.

        Examples:
            >>> from anonipy.anonymize.regex import RegexMapping
            >>> regex_mapping = RegexMapping()
            >>> regex_mapping["string"]
            ".*"

        Args:
            regex_type: The type of the entity.

        Returns:
            The regex for the given type.

        """
        return self.regex_mapping[regex_type]


regex_mapping = RegexMapping()
"""The shorthand to the `RegexMapping` instance."""



"""Module containing the `constants`.

The `constants` module provides a set of predefined constants used in the package.
These include supported languages, types of entities, and date transformation
variants.

Classes:
    LANGUAGES: Predefined supported languages.
    ENTITY_TYPES: Predefined types of entities.
    DATE_TRANSFORM_VARIANTS: Predefined types of the date transformation variants.

"""

from typing import List


class LANGUAGES:
    """The main anonipy supported languages.

    Attributes:
        DUTCH (Tuple[Literal["nl"], Literal["Dutch"]]): The Dutch language.
        ENGLISH (Tuple[Literal["en"], Literal["English"]]): The English language.
        FRENCH (Tuple[Literal["fr"], Literal["French"]]): The French language.
        GERMAN (Tuple[Literal["de"], Literal["German"]]): The German language.
        GREEK (Tuple[Literal["el"], Literal["Greek"]]): The Greek language.
        ITALIAN (Tuple[Literal["it"], Literal["Italian"]]): The Italian language.
        SLOVENIAN (Tuple[Literal["sl"], Literal["Slovenian"]]): The Slovenian language.
        SPANISH (Tuple[Literal["es"], Literal["Spanish"]]): The Spanish language.
        UKRAINIAN (Tuple[Literal["uk"], Literal["Ukrainian"]]): The Ukrainian language.

    Methods:
        supported_languages():
            Return a list of supported languages.

    """

    DUTCH = ("nl", "Dutch")
    ENGLISH = ("en", "English")
    FRENCH = ("fr", "French")
    GERMAN = ("de", "German")
    GREEK = ("el", "Greek")
    ITALIAN = ("it", "Italian")
    SLOVENIAN = ("sl", "Slovenian")
    SPANISH = ("es", "Spanish")
    UKRAINIAN = ("uk", "Ukrainian")

    @classmethod
    def supported_languages(cls) -> List[str]:
        """Return a list of supported languages.

        Returns:
            The list of supported languages.

        """
        languages = [
            cls.DUTCH[0],
            cls.ENGLISH[0],
            cls.FRENCH[0],
            cls.GERMAN[0],
            cls.GREEK[0],
            cls.ITALIAN[0],
            cls.SLOVENIAN[0],
            cls.SPANISH[0],
            cls.UKRAINIAN[0],
        ]
        return languages


class ENTITY_TYPES:
    """The anonipy supported entity types.

    Attributes:
        CUSTOM (Literal["custom"]): The custom entity type.
        STRING (Literal["string"]): The string entity type.
        INTEGER (Literal["integer"]): The integer entity type.
        FLOAT (Literal["float"]): The float entity type.
        DATE (Literal["date"]): The date entity type.
        EMAIL (Literal["email"]): The email entity type.
        WEBSITE_URL (Literal["website_url"]): The website url entity type.
        PHONE_NUMBER (Literal["phone_number"]): The phone number entity type.

    """

    CUSTOM = "custom"
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    DATE = "date"
    EMAIL = "email"
    WEBSITE_URL = "website_url"
    PHONE_NUMBER = "phone_number"


class DATE_TRANSFORM_VARIANTS:
    """The supported date transform variants.

    Attributes:
        FIRST_DAY_OF_THE_MONTH (Literal["FIRST_DAY_OF_THE_MONTH"]): The first day of the month.
        LAST_DAY_OF_THE_MONTH (Literal["LAST_DAY_OF_THE_MONTH"]): The last day of the month.
        MIDDLE_OF_THE_MONTH (Literal["MIDDLE_OF_THE_MONTH"]): The middle of the month.
        MIDDLE_OF_THE_YEAR (Literal["MIDDLE_OF_THE_YEAR"]): The middle of the year.
        RANDOM (Literal["RANDOM"]): A random date.

    Methods:
        values():
            Return a list of all possible date transform variants.
        is_valid(value):
            Check if the value is a valid date variant.

    """

    FIRST_DAY_OF_THE_MONTH = "FIRST_DAY_OF_THE_MONTH"
    LAST_DAY_OF_THE_MONTH = "LAST_DAY_OF_THE_MONTH"
    MIDDLE_OF_THE_MONTH = "MIDDLE_OF_THE_MONTH"
    MIDDLE_OF_THE_YEAR = "MIDDLE_OF_THE_YEAR"
    RANDOM = "RANDOM"

    @classmethod
    def values(cls) -> List[str]:
        """Return a list of all possible date transform variants.

        Returns:
            The list of all possible variants.

        """
        return [
            cls.FIRST_DAY_OF_THE_MONTH,
            cls.LAST_DAY_OF_THE_MONTH,
            cls.MIDDLE_OF_THE_MONTH,
            cls.MIDDLE_OF_THE_YEAR,
            cls.RANDOM,
        ]

    @classmethod
    def is_valid(cls, value: str) -> bool:
        """Check if the value is a valid date variant.

        Args:
            value: The value to check.

        Returns:
            `True` if the value is a valid date variant, `False` otherwise.

        """
        return value in cls.values()



import random
import warnings
from .interface import GeneratorInterface
from ...definitions import Entity

# =====================================
# Main class
# =====================================


class NumberGenerator(GeneratorInterface):
    """The class representing the number generator.

    Examples:
        >>> from anonipy.anonymize.generators import NumberGenerator
        >>> generator = NumberGenerator()
        >>> generator.generate(entity)

    Methods:
        generate(self, entity):
            Generates a substitute for the numeric entity.

    """

    def __init__(self, *args, **kwargs):
        """Initializes the number generator.

        Examples:
            >>> from anonipy.anonymize.generators import NumberGenerator
            >>> generator = NumberGenerator()

        """

        super().__init__(*args, **kwargs)

    def generate(self, entity: Entity, *args, **kwargs) -> str:
        """Generates the substitute for the numeric entity.

        Examples:
            >>> from anonipy.anonymize.generators import NumberGenerator
            >>> generator = NumberGenerator()
            >>> generator.generate(entity)
            "1234567890"

        Args:
            entity: The numeric entity to generate the numeric substitute.

        Returns:
            The generated numeric substitute.

        Raises:
            ValueError: If the entity type is not `integer`, `float`, `phone_number` or `custom`.

        """

        if entity.type in ["custom"]:
            warnings.warn(
                "The entity type is `custom`. Make sure the generator is returning appropriate values."
            )
        elif entity.type not in ["integer", "float", "phone_number"]:
            raise ValueError(
                "The entity type must be `integer`, `float`, `phone_number` or `custom` to generate numbers."
            )
        return "".join(
            [str(random.randint(0, 9)) if d.isdigit() else d for d in entity.text]
        )



"""`Anonipy` is a text anonymization package.

The `anonipy` package provides utilities for data anonymization. It provides
a set of modules and utilities for (1) identifying relevant information
that needs to be anonymized, (2) generating substitutes for the identified
information, and (3) strategies for anonymizing the identified information.

Modules:
    anonymize: The module containing the anonymization submodules and utility.
    utils: The module containing utility classes and functions.
    definitions: The module containing predefined types used across the package.
    constants: The module containing the predefined constants used across the package.

"""

__version__ = "0.6.0"

from . import anonymize
from . import utils
from . import definitions
from . import constants



import pytest

from anonipy.definitions import Entity
from anonipy.anonymize.strategies import (
    RedactionStrategy,
    MaskingStrategy,
    PseudonymizationStrategy,
)

# =====================================
# Helper functions
# =====================================

TEST_TEXT = "Test this string, and this test too!"
TEST_ENTITIES = [
    Entity(text="Test", label="test", start_index=0, end_index=4),
    Entity(text="string", label="type", start_index=10, end_index=16),
    Entity(text="test", label="test", start_index=27, end_index=31),
]


def anonymization_mapping(text, entity):
    if entity.label == "test":
        return "[TEST]"
    elif entity.label == "type":
        return "[TYPE]"
    return "[REDACTED]"


@pytest.fixture
def redaction_strategy():
    return RedactionStrategy()


@pytest.fixture
def masking_strategy():
    return MaskingStrategy()


@pytest.fixture
def pseudonymization_strategy():
    return PseudonymizationStrategy(mapping=anonymization_mapping)


# =====================================
# Test Redaction Strategy
# =====================================


def test_redaction_strategy_init(redaction_strategy):
    assert redaction_strategy.__class__ == RedactionStrategy


def test_redaction_strategy_has_methods(redaction_strategy):
    assert hasattr(redaction_strategy, "anonymize")


@pytest.mark.parametrize(
    "substitute_label, expected_label",
    [
        (None, "[REDACTED]"),
        ("[TEST]", "[TEST]"),
    ],
)
def test_redaction_strategy_inputs(substitute_label, expected_label):
    strategy = RedactionStrategy(substitute_label=substitute_label)
    assert strategy.substitute_label == expected_label


@pytest.mark.parametrize(
    "substitute_label, expected_text, expected_replacements",
    [
        (
            None,
            "[REDACTED] this [REDACTED], and this [REDACTED] too!",
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "[REDACTED]",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "[REDACTED]",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "[REDACTED]",
                },
            ],
        ),
        (
            "[TEST]",
            "[TEST] this [TEST], and this [TEST] too!",
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "[TEST]",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "[TEST]",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "[TEST]",
                },
            ],
        ),
    ],
)
def test_redaction_strategy_anonymize(
    substitute_label, expected_text, expected_replacements
):
    strategy = RedactionStrategy(substitute_label=substitute_label)
    anonymized_text, replacements = strategy.anonymize(TEST_TEXT, TEST_ENTITIES)
    assert anonymized_text == expected_text
    assert replacements == expected_replacements


# =====================================
# Test Masking Strategy
# =====================================


def test_masking_strategy_init(masking_strategy):
    assert masking_strategy.__class__ == MaskingStrategy


def test_masking_strategy_methods(masking_strategy):
    assert hasattr(masking_strategy, "anonymize")


@pytest.mark.parametrize(
    "substitute_label, expected_label",
    [
        (None, "*"),
        ("A", "A"),
    ],
)
def test_masking_strategy_inputs(substitute_label, expected_label):
    strategy = MaskingStrategy(substitute_label=substitute_label)
    assert strategy.substitute_label == expected_label


@pytest.mark.parametrize(
    "substitute_label, expected_text, expected_replacements",
    [
        (
            None,
            "**** this ******, and this **** too!",
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "****",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "******",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "****",
                },
            ],
        ),
        (
            "A",
            "AAAA this AAAAAA, and this AAAA too!",
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "AAAA",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "AAAAAA",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "AAAA",
                },
            ],
        ),
    ],
)
def test_masking_strategy_anonymize(
    substitute_label, expected_text, expected_replacements
):
    strategy = MaskingStrategy(substitute_label=substitute_label)
    anonymized_text, replacements = strategy.anonymize(TEST_TEXT, TEST_ENTITIES)
    assert anonymized_text == expected_text
    assert replacements == expected_replacements


# =====================================
# Test Pseudonymization Strategy
# =====================================


def test_pseudonymization_strategy_init():
    with pytest.raises(TypeError):
        PseudonymizationStrategy()


def test_pseudonymization_strategy_init_inputs(pseudonymization_strategy):
    assert pseudonymization_strategy.__class__ == PseudonymizationStrategy


def test_pseudonymization_strategy_methods(pseudonymization_strategy):
    assert hasattr(pseudonymization_strategy, "anonymize")


def test_pseudonymization_strategy_anonymize_inputs(pseudonymization_strategy):
    anonymized_text, replacements = pseudonymization_strategy.anonymize(
        TEST_TEXT, TEST_ENTITIES
    )
    assert anonymized_text == "[TEST] this [TYPE], and this [TEST] too!"
    assert replacements == [
        {
            "original_text": "Test",
            "label": "test",
            "start_index": 0,
            "end_index": 4,
            "anonymized_text": "[TEST]",
        },
        {
            "original_text": "string",
            "label": "type",
            "start_index": 10,
            "end_index": 16,
            "anonymized_text": "[TYPE]",
        },
        {
            "original_text": "test",
            "label": "test",
            "start_index": 27,
            "end_index": 31,
            "anonymized_text": "[TEST]",
        },
    ]



from typing import List, Tuple
import itertools

from spacy import displacy
from spacy.tokens import Doc

from ...definitions import Entity
from ...utils.colors import get_label_color
from ..helpers import merge_entities

from .interface import ExtractorInterface


# ===============================================
# Extractor class
# ===============================================


class MultiExtractor:
    """The class representing the multi extractor.

    Examples:
        >>> from anonipy.constants import LANGUAGES
        >>> from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
        >>> extractors = [
        >>>     NERExtractor(ner_labels, lang=LANGUAGES.ENGLISH),
        >>>     PatternExtractor(pattern_labels, lang=LANGUAGES.ENGLISH),
        >>> ]
        >>> extractor = MultiExtractor(extractors)
        >>> extractor("John Doe is a 19 year old software engineer.", detect_repeats=False)
        [(Doc, [Entity]), (Doc, [Entity])], [Entity]

    Attributes:
        extractors (List[ExtractorInterface]):
            The list of extractors to use.

    Methods:
        __call__(self, text):
            Extract the entities fron the text using the provided extractors.
        display(self, doc):
            Display the entities extracted from the text document.

    """

    def __init__(self, extractors: List[ExtractorInterface]):
        """Initialize the multi extractor.

        Examples:
            >>> from anonipy.constants import LANGUAGES
            >>> from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
            >>> extractors = [
            >>>     NERExtractor(ner_labels, lang=LANGUAGES.ENGLISH),
            >>>     PatternExtractor(pattern_labels, lang=LANGUAGES.ENGLISH),
            >>> ]
            >>> extractor = MultiExtractor(extractors)
            MultiExtractor()

        Args:
            extractors: The list of extractors to use.

        """
        if len(extractors) == 0:
            raise ValueError("At least one extractor must be provided.")
        if not all(isinstance(e, ExtractorInterface) for e in extractors):
            raise ValueError("All extractors must be instances of ExtractorInterface.")

        self.extractors = extractors

    def __call__(
        self, text: str, detect_repeats: bool = False
    ) -> Tuple[List[Tuple[Doc, List[Entity]]], List[Entity]]:
        """Extract the entities fron the text using the provided extractors.

        Examples:
            >>> extractor("John Doe is a 19 year old software engineer.", detect_repeats=False)
            [(Doc, [Entity]), (Doc, [Entity])], [Entity]

        Args:
            text: The text to extract entities from.
            detect_repeats: Whether to check text again for repeated entities.

        Returns:
            The list of extractor outputs containing the tuple (spacy document, extracted entities).
            The list of joint entities.

        """

        extractor_outputs = [e(text, detect_repeats) for e in self.extractors]
        joint_entities = merge_entities(extractor_outputs)

        return extractor_outputs, joint_entities

    def display(self, doc: Doc, page: bool = False, jupyter: bool = None) -> str:
        """Display the entities in the text.

        Examples:
            >>> extractor_outputs, entities = extractor("John Doe is a 19 year old software engineer.")
            >>> extractor.display(extractor_outputs[0][0])
            HTML

        Args:
            doc: The spacy doc to display.
            page: Whether to display the doc in a web browser.
            jupyter: Whether to display the doc in a jupyter notebook.

        Returns:
            The HTML representation of the document and the extracted entities.

        """

        labels = list(
            itertools.chain.from_iterable([e.labels for e in self.extractors])
        )
        options = {"colors": {l["label"]: get_label_color(l["label"]) for l in labels}}
        return displacy.render(
            doc, style="ent", options=options, page=page, jupyter=jupyter
        )



"""Module containing the `pipeline`.

The `pipeline` module provides a class for anonymizing files using a pipeline of
extractors and strategies.

Classes:
    Pipeline: The class representing the anonymization pipeline.

"""

import os
import warnings
from typing import Union, List

from .extractors import ExtractorInterface, MultiExtractor
from .strategies import StrategyInterface
from ..utils.file_system import open_file, write_file


# =====================================
# Pipeline class
# =====================================


class Pipeline:
    """A class for anonymizing files using a pipeline of extractors and strategies.

    Examples:
        >>> from anonipy.anonymize.pipeline import Pipeline
        >>> extractor = NERExtractor(labels, lang=LANGUAGES.ENGLISH)
        >>> strategy = RedactionStrategy()
        >>> pipeline = Pipeline(extractor, strategy)
        >>> pipeline.anonymize("/path/to/input_dir", "/path/to/output_dir", flatten=True)

    Attributes:
        extractor (ExtractorInterface, MultiExtractor, List[ExtractorInterface]): The extractor to use for entity extraction.
        strategy (StrategyInterface): The strategy to use for anonymization.

    Methods:
        anonymize(input_dir, output_dir, flatten=False):
            Anonymize files in the input directory and save the anonymized files to the output directory.

    """

    def __init__(
        self,
        extractor: Union[ExtractorInterface, MultiExtractor, List[ExtractorInterface]],
        strategy: StrategyInterface,
    ):
        """Initialize the pipeline.

        Examples:
            >>> from anonipy.anonymize.pipeline import Pipeline
            >>> extractor = NERExtractor(labels, lang=LANGUAGES.ENGLISH)
            >>> strategy = RedactionStrategy()
            >>> pipeline = Pipeline(extractor, strategy)

        Args:
            extractor: The extractor to use for entity extraction.
            strategy: The strategy to use for anonymization.

        """

        if isinstance(extractor, ExtractorInterface) or isinstance(
            extractor, MultiExtractor
        ):
            self.extractor = extractor
        elif isinstance(extractor, list):
            self.extractor = MultiExtractor(extractor)
        else:
            raise ValueError(
                "Extractor must be an ExtractorInterface or a list of ExtractorInterface."
            )

        if not isinstance(strategy, StrategyInterface):
            raise ValueError("Strategy must be a StrategyInterface.")

        self.strategy = strategy

    def anonymize(self, input_dir: str, output_dir: str, flatten: bool = False) -> dict:
        """Anonymize files in the input directory and save the anonymized files to the output directory.

        Args:
            input_dir: The path to the input directory containing files to be anonymized.
            output_dir: The path to the output directory where anonymized files will be saved.
            flatten: Whether to flatten the output directory structure. Defaults to False.

        Raises:
            ValueError: If the input directory does not exist or if the input and output directories are the same.

        Returns:
            A dictionary mapping the original file paths to the anonymized file paths.

        """

        if not os.path.exists(input_dir):
            raise ValueError(f"Input directory '{input_dir}' does not exist.")

        if os.path.abspath(input_dir) == os.path.abspath(output_dir):
            raise ValueError("Input and output directories cannot be the same.")

        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        anonymized_files_count = 1
        file_name_mapping = {}

        for root, _, files in os.walk(input_dir):
            for file_name in files:
                file_path = os.path.join(root, file_name)

                try:
                    anonymized_text = self._anonymize_file(file_path)
                    if anonymized_text is None:
                        continue
                except Exception as e:
                    warnings.warn(f"Problems while processing file {file_path}: {e}")
                    continue

                _, ext = os.path.splitext(file_name)
                output_file_name = f"file{anonymized_files_count}_anony{ext}"
                anonymized_files_count += 1

                relative_path = os.path.relpath(file_path, input_dir)

                if flatten:
                    output_file_path = os.path.join(output_dir, output_file_name)
                else:
                    output_file_path = os.path.join(
                        output_dir, os.path.dirname(relative_path), output_file_name
                    )
                    os.makedirs(os.path.dirname(output_file_path), exist_ok=True)

                write_file(anonymized_text, output_file_path)

                file_path_before = os.path.join(
                    input_dir.split(os.sep)[-1], relative_path
                )
                file_path_after = os.path.relpath(output_file_path, output_dir)
                file_name_mapping[file_path_before] = os.path.join(
                    output_dir.split(os.sep)[-1], file_path_after
                )

        return file_name_mapping

    def _anonymize_file(self, file_path: str) -> Union[str, None]:
        """Anonymize a single file.

        Args:
            file_path: The path to the file to be anonymized.

        Returns:
            The anonymized text or None if the file is empty or if entity extraction fails.

        """

        original_text = open_file(file_path)
        if original_text is None or not original_text.strip():
            warnings.warn(
                f"Skipping file {file_path}: Failed to read or file is empty."
            )
            return None

        _, entities = self.extractor(original_text)

        if not entities:
            warnings.warn(
                f"Skipping file {file_path}: Entity extraction returned None."
            )
            return None

        anonymized_text, _ = self.strategy.anonymize(original_text, entities)

        return anonymized_text



"""The module containing the `utils`.

The `utils` module provides a set of utilities used in the package.

Modules:
    regex: The module containing the regex utilities and functions.
    file_system: The module containing the file system utilities and functions.
    language_detector: The module containing the language detector.

"""

from ..utils import regex
from . import file_system
from . import language_detector

__all__ = ["regex", "file_system", "language_detector"]



import warnings

import pytest
import torch
from transformers import logging

from anonipy.definitions import Entity
from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
from anonipy.constants import LANGUAGES
from anonipy.anonymize.helpers import filter_entities

# disable transformers logging
logging.set_verbosity_error()

# =====================================
# Helper functions
# =====================================

TEST_ORIGINAL_TEXT = """\
Medical Record

Patient Name: John Doe
Date of Birth: 15-01-1985
Date of Examination: 20-05-2024
Social Security Number: 123-45-6789

Examination Procedure:
John Doe underwent a routine physical examination. The procedure included measuring vital signs (blood pressure, heart rate, temperature), a comprehensive blood panel, and a cardiovascular stress test. The patient also reported occasional headaches and dizziness, prompting a neurological assessment and an MRI scan to rule out any underlying issues.

Medication Prescribed:

Ibuprofen 200 mg: Take one tablet every 6-8 hours as needed for headache and pain relief.
Lisinopril 10 mg: Take one tablet daily to manage high blood pressure.
Next Examination Date:
15-11-2024

The examination took place on 20-05-2024. John Doe was prescribed Ibuprofen 200 mg and Lisinopril 10 mg.
"""

TEST_NER_ENTITIES = [
    Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        type="string",
    ),
    Entity(
        text="15-01-1985",
        label="date of birth",
        start_index=54,
        end_index=64,
        type="date",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type="date",
    ),
    Entity(
        text="123-45-6789",
        label="social security number",
        start_index=121,
        end_index=132,
        type="custom",
        regex="[0-9]{3}-[0-9]{2}-[0-9]{4}",
    ),
    Entity(
        text="John Doe",
        label="name",
        start_index=157,
        end_index=165,
        type="string",
    ),
    Entity(
        text="15-11-2024",
        label="date",
        start_index=717,
        end_index=727,
        type="date",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=759,
        end_index=769,
        type="date",
    ),
]
TEST_REPEATS_ENTITIES = [
    Entity(
        text="John Doe",
        label="name",
        start_index=771,
        end_index=779,
        type="string",
    ),
]
TEST_PATTERN_ENTITIES = [
    Entity(
        text="15-01-1985",
        label="date",
        start_index=54,
        end_index=64,
        type="date",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type="date",
    ),
    Entity(
        text="blood pressure, heart rate, temperature",
        label="symptoms",
        start_index=254,
        end_index=293,
        type="string",
        regex="\\((.*)\\)",
    ),
    Entity(
        text="Ibuprofen 200 mg",
        label="medicine",
        start_index=533,
        end_index=549,
        type="string",
    ),
    Entity(
        text="Lisinopril 10 mg",
        label="medicine",
        start_index=623,
        end_index=639,
        type="string",
    ),
    Entity(
        text="15-11-2024",
        label="date",
        start_index=717,
        end_index=727,
        type="date",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=759,
        end_index=769,
        type="date",
    ),
    Entity(
        text="Ibuprofen 200 mg",
        label="medicine",
        start_index=795,
        end_index=811,
        type="string",
    ),
    Entity(
        text="Lisinopril 10 mg",
        label="medicine",
        start_index=816,
        end_index=832,
        type="string",
    ),
]
TEST_PATTERN_DETECT_REPEATS = [
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type="date",
        regex=r"Date of Examination: (.*)",
    ),
    # Repeated entity
    Entity(
        text="20-05-2024",
        label="date",
        start_index=759,
        end_index=769,
        type="date",
        regex=r"Date of Examination: (.*)",
    ),
]
TEST_MULTI_REPEATS = [
    Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        type="string",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type="date",
        regex=r"Date of Examination: (.*)",
    ),
    Entity(
        text="John Doe",
        label="name",
        start_index=157,
        end_index=165,
        type="string",
    ),
    # Repeated entities
    Entity(
        text="20-05-2024",
        label="date",
        start_index=759,
        end_index=769,
        type="date",
        regex=r"Date of Examination: (.*)",
    ),
    Entity(
        text="John Doe",
        label="name",
        start_index=771,
        end_index=779,
        type="string",
    ),
]


@pytest.fixture(autouse=True)
def suppress_warnings():
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=ResourceWarning)


@pytest.fixture(scope="module")
def ner_extractor():
    labels = [
        {"label": "name", "type": "string"},
        {
            "label": "social security number",
            "type": "custom",
            "regex": "[0-9]{3}-[0-9]{2}-[0-9]{4}",
        },
        {"label": "date of birth", "type": "date"},
        {"label": "date", "type": "date"},
    ]
    return NERExtractor(labels=labels, lang=LANGUAGES.ENGLISH)


@pytest.fixture(scope="module")
def pattern_extractor():
    labels = [
        {
            "label": "symptoms",
            "type": "string",
            "regex": r"\((.*)\)",  # symptoms are enclosed in parentheses
        },
        {
            "label": "medicine",
            "type": "string",
            "pattern": [[{"IS_ALPHA": True}, {"LIKE_NUM": True}, {"LOWER": "mg"}]],
        },
        {
            "label": "date",
            "type": "date",
            "pattern": [  # represent the date as a sequence of digits using spacy
                [
                    {"SHAPE": "dd"},
                    {"TEXT": "-"},
                    {"SHAPE": "dd"},
                    {"TEXT": "-"},
                    {"SHAPE": "dddd"},
                ]
            ],
        },
    ]
    return PatternExtractor(labels=labels, lang=LANGUAGES.ENGLISH)


@pytest.fixture(scope="module")
def multi_extractor(ner_extractor, pattern_extractor):
    return MultiExtractor([ner_extractor, pattern_extractor])


def test_ner_extractor_init():
    with pytest.raises(TypeError):
        NERExtractor()


def test_ner_extractor_init_inputs(ner_extractor):
    extractor = NERExtractor(
        labels=ner_extractor.labels, lang=LANGUAGES.ENGLISH, score_th=0.5
    )
    assert isinstance(extractor, NERExtractor)


def test_ner_extractor_init_gpu(ner_extractor):
    if torch.cuda.is_available():
        extractor = NERExtractor(
            labels=ner_extractor.labels,
            lang=LANGUAGES.ENGLISH,
            score_th=0.5,
            use_gpu=True,
        )
        assert isinstance(extractor, NERExtractor)


def test_ner_extractor_methods(ner_extractor):
    assert hasattr(ner_extractor, "__call__")
    assert hasattr(ner_extractor, "display")


def test_ner_extractor_extract_default_params(ner_extractor):
    _, entities = ner_extractor(TEST_ORIGINAL_TEXT)
    for p_entity, t_entity in zip(entities, TEST_NER_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_ner_extractor_extract_default_params_input():
    extractor = NERExtractor(
        labels=[
            {"label": "name", "type": "string"},
            {
                "label": "social security number",
                "type": "custom",
                "regex": "[0-9]{3}-[0-9]{2}-[0-9]{4}",
            },
            {"label": "date of birth", "type": "date"},
            {"label": "date", "type": "date"},
        ],
        lang=LANGUAGES.ENGLISH,
        gliner_model="urchade/gliner_multi_pii-v1",
        score_th=0.5,
    )
    _, entities = extractor(TEST_ORIGINAL_TEXT)
    for p_entity, t_entity in zip(entities, TEST_NER_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_ner_extractor_extract_custom_params_input():
    extractor = NERExtractor(
        labels=[
            {"label": "name", "type": "string"},
            {
                "label": "social security number",
                "type": "custom",
                "regex": "[0-9]{3}-[0-9]{2}-[0-9]{4}",
            },
            {"label": "date of birth", "type": "date"},
            {"label": "date", "type": "date"},
        ],
        lang=LANGUAGES.ENGLISH,
        gliner_model="E3-JSI/gliner-multi-pii-domains-v1",
        score_th=0.5,
    )
    _, entities = extractor(TEST_ORIGINAL_TEXT)
    for p_entity, t_entity in zip(entities, TEST_NER_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_ner_extractor_detect_repeats_true(ner_extractor):
    _, entities = ner_extractor(TEST_ORIGINAL_TEXT, detect_repeats=True)
    expected_entities = TEST_NER_ENTITIES + TEST_REPEATS_ENTITIES
    for p_entity, t_entity in zip(entities, expected_entities):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_pattern_extractor_init():
    with pytest.raises(TypeError):
        PatternExtractor()


def test_pattern_extractor_init_inputs(pattern_extractor):
    assert isinstance(pattern_extractor, PatternExtractor)


def test_pattern_extractor_methods(pattern_extractor):
    assert hasattr(pattern_extractor, "__call__")
    assert hasattr(pattern_extractor, "display")


def test_pattern_extractor_extract_default(pattern_extractor):
    doc, entities = pattern_extractor(TEST_ORIGINAL_TEXT)
    for p_entity, t_entity in zip(entities, TEST_PATTERN_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score == 1.0


def test_pattern_extractor_detect_repeats_false():
    extractor = PatternExtractor(
        labels=[
            {
                "label": "date",
                "type": "date",
                "regex": r"Date of Examination: (.*)",
            }
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(TEST_ORIGINAL_TEXT, detect_repeats=False)
    excepted_entity = TEST_PATTERN_DETECT_REPEATS[0]
    assert len(entities) == 1
    assert excepted_entity.text == entities[0].text
    assert excepted_entity.label == entities[0].label
    assert excepted_entity.start_index == entities[0].start_index
    assert excepted_entity.end_index == entities[0].end_index
    assert excepted_entity.type == entities[0].type
    assert excepted_entity.regex == entities[0].regex
    assert excepted_entity.score >= 0.5


def test_pattern_extractor_detect_repeats_true():
    extractor = PatternExtractor(
        labels=[
            {
                "label": "date",
                "type": "date",
                "regex": r"Date of Examination: (.*)",
            }
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(TEST_ORIGINAL_TEXT, detect_repeats=True)
    for p_entity, t_entity in zip(entities, TEST_PATTERN_DETECT_REPEATS):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_pattern_extractor_numbers_ssn_regex():
    """Test PatternExtractor for identifying Social Security Numbers using regex"""
    extractor = PatternExtractor(
        labels=[
            {
                "label": "social security number",
                "type": "custom",
                "regex": r"([0-9]{3}-[0-9]{2}-[0-9]{4})",
            }
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(TEST_ORIGINAL_TEXT)
    assert len(entities) == 1
    assert entities[0].text == "123-45-6789"
    assert entities[0].label == "social security number"
    assert entities[0].start_index == 121
    assert entities[0].end_index == 132
    assert entities[0].type == "custom"
    assert entities[0].score == 1.0


def test_pattern_extractor_numbers_dosage():
    """Test PatternExtractor for identifying numeric dosages using patterns"""
    extractor = PatternExtractor(
        labels=[
            {
                "label": "dosage",
                "type": "string",
                "pattern": [[{"LIKE_NUM": True}, {"LOWER": "mg"}]],
            }
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(TEST_ORIGINAL_TEXT)
    # Should find "200 mg" and "10 mg" (appears twice each in the text)
    assert len(entities) >= 2
    dosages = [entity.text for entity in entities]
    assert "200 mg" in dosages
    assert "10 mg" in dosages
    for entity in entities:
        assert entity.label == "dosage"
        assert entity.type == "string"
        assert entity.score == 1.0


def test_pattern_extractor_numbers_phone():
    """Test PatternExtractor for identifying phone numbers using regex"""
    test_text = "Contact: 555-123-4567 or call (555) 987-6543 for more info."
    extractor = PatternExtractor(
        labels=[
            {
                "label": "phone",
                "type": "custom",
                "regex": r"(\d{3}-\d{3}-\d{4})",
            }
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(test_text)
    assert len(entities) == 1
    assert entities[0].text == "555-123-4567"
    assert entities[0].label == "phone"
    assert entities[0].type == "custom"
    assert entities[0].score == 1.0


def test_pattern_extractor_numbers_general():
    """Test PatternExtractor for identifying general numeric values"""
    test_text = "The patient weighs 150 pounds and is 72 inches tall."
    extractor = PatternExtractor(
        labels=[
            {
                "label": "measurement",
                "type": "string",
                "pattern": [[{"LIKE_NUM": True}, {"IS_ALPHA": True}]],
            }
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(test_text)
    assert len(entities) == 2
    measurements = [entity.text for entity in entities]
    assert "150 pounds" in measurements
    assert "72 inches" in measurements
    for entity in entities:
        assert entity.label == "measurement"
        assert entity.type == "string"
        assert entity.score == 1.0


def test_pattern_extractor_numbers_mixed_formats():
    """Test PatternExtractor for identifying numbers in various formats"""
    test_text = "Account: 12345, Balance: $1,234.56, Code: ABC-123"
    extractor = PatternExtractor(
        labels=[
            {
                "label": "account_number",
                "type": "custom",
                "regex": r"Account: (\d+)",
            },
            {
                "label": "amount",
                "type": "custom",
                "regex": r"Balance: (\$[\d,]+\.?\d*)",
            },
        ],
        lang=LANGUAGES.ENGLISH,
    )
    _, entities = extractor(test_text)
    assert len(entities) == 2

    # Check account number
    account_entity = [e for e in entities if e.label == "account_number"][0]
    assert account_entity.text == "12345"
    assert account_entity.type == "custom"

    # Check amount
    amount_entity = [e for e in entities if e.label == "amount"][0]
    assert amount_entity.text == "$1,234.56"
    assert amount_entity.type == "custom"


def test_multi_extractor_init():
    with pytest.raises(TypeError):
        MultiExtractor()


def test_multi_extractor_init_inputs(multi_extractor):
    assert isinstance(multi_extractor, MultiExtractor)


def test_multi_extractor_methods(multi_extractor):
    assert hasattr(multi_extractor, "__call__")
    assert hasattr(multi_extractor, "display")


def test_multi_extractor_extract_default(multi_extractor):
    extractor_outputs, joint_entities = multi_extractor(TEST_ORIGINAL_TEXT)

    # check the performance of the first extractor
    for p_entity, t_entity in zip(extractor_outputs[0][1], TEST_NER_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5

    # check the performance of the second extractor
    for p_entity, t_entity in zip(extractor_outputs[1][1], TEST_PATTERN_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score == 1.0

    # check the performance of the joint entities generation
    for p_entity, t_entity in zip(
        joint_entities,
        filter_entities(TEST_NER_ENTITIES + TEST_REPEATS_ENTITIES + TEST_PATTERN_ENTITIES),
    ):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_multi_extractor_extract_empty_extractor_list():
    with pytest.raises(ValueError):
        MultiExtractor([])


def test_multi_extractor_extract_invalid_extractor_list():
    with pytest.raises(ValueError):
        MultiExtractor([NERExtractor(labels=[], lang=LANGUAGES.ENGLISH), "invalid"])


def test_multi_extractor_extract_single_extractor_ner(multi_extractor):
    extractor = MultiExtractor([multi_extractor.extractors[0]])
    extractor_outputs, joint_entities = extractor(TEST_ORIGINAL_TEXT)

    # check the performance of the extractor
    for p_entity, t_entity in zip(extractor_outputs[0][1], TEST_NER_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5

    # check the performance of the joint entities generation
    for p_entity, t_entity in zip(joint_entities, TEST_NER_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_multi_extractor_extract_single_extractor_pattern(multi_extractor):
    extractor = MultiExtractor([multi_extractor.extractors[1]])
    extractor_outputs, joint_entities = extractor(TEST_ORIGINAL_TEXT)

    # check the performance of the extractor
    for p_entity, t_entity in zip(extractor_outputs[0][1], TEST_PATTERN_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5

    # check the performance of the joint entities generation
    for p_entity, t_entity in zip(joint_entities, TEST_PATTERN_ENTITIES):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_multi_extractor_detect_repeats_false():
    extractors = [
        NERExtractor(
            labels=[
                {"label": "name", "type": "string"},
            ]
        ),
        PatternExtractor(
            labels=[
                {
                    "label": "date",
                    "type": "date",
                    "regex": r"Date of Examination: (.*)",
                },
            ]
        ),
    ]
    extractor = MultiExtractor(extractors)
    _, joint_entities = extractor(TEST_ORIGINAL_TEXT, detect_repeats=False)
    for p_entity, t_entity in zip(joint_entities, TEST_MULTI_REPEATS[:3]):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5


def test_multi_extractor_detect_repeats_true():
    extractors = [
        NERExtractor(
            labels=[
                {"label": "name", "type": "string"},
            ]
        ),
        PatternExtractor(
            labels=[
                {
                    "label": "date",
                    "type": "date",
                    "regex": r"Date of Examination: (.*)",
                },
            ]
        ),
    ]
    extractor = MultiExtractor(extractors)
    _, joint_entities = extractor(TEST_ORIGINAL_TEXT, detect_repeats=True)
    for p_entity, t_entity in zip(joint_entities, TEST_MULTI_REPEATS):
        assert p_entity.text == t_entity.text
        assert p_entity.label == t_entity.label
        assert p_entity.start_index == t_entity.start_index
        assert p_entity.end_index == t_entity.end_index
        assert p_entity.type == t_entity.type
        assert p_entity.regex == t_entity.regex
        assert p_entity.score >= 0.5



"""Module containing the `strategies`.

The `strategies` module provides a set of strategies used to anonymize the
identified vulnerable data.

Classes:
    RedactionStrategy: The class representing the redaction strategy.
    MaskingStrategy: The class representing the masking strategy.
    PseudonymizationStrategy: The class representing the pseudonymization strategy.

"""

from .interface import StrategyInterface
from .masking import MaskingStrategy
from .redaction import RedactionStrategy
from .pseudonymization import PseudonymizationStrategy


__all__ = [
    "StrategyInterface",
    "MaskingStrategy",
    "RedactionStrategy",
    "PseudonymizationStrategy",
]



import random

random.seed(42)

# ===============================================
# GLOBAL COLOR MAPPING
# ===============================================

GLOBAL_COLOR_MAPPING = {}

# ===============================================
# Color generators and functions
# ===============================================


def generate_random_color():
    return "#" + "".join([random.choice("0123456789ABCDEF") for j in range(6)])


def get_label_color(label):
    if label in GLOBAL_COLOR_MAPPING:
        return GLOBAL_COLOR_MAPPING[label]
    else:
        color = generate_random_color()
        GLOBAL_COLOR_MAPPING[label] = color
        return color
